@@ -1,4 +1,4 @@
-// Copyright 2012-2019 The NATS Authors
+// Copyright 2012-2022 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -21,8 +21,10 @@ import (
 	"io"
 	"math"
 	"net"
+	"net/url"
 	"reflect"
 	"regexp"
+	"runtime"
 	"strings"
 	"sync"
 	"sync/atomic"
@@ -32,17 +34,23 @@ import (
 	"crypto/rand"
 	"crypto/tls"
 
+	"github.com/nats-io/jwt/v2"
 	"github.com/nats-io/nats.go"
+	"github.com/nats-io/nkeys"
 )
 
 type serverInfo struct {
-	ID           string `json:"server_id"`
-	Host         string `json:"host"`
-	Port         uint   `json:"port"`
-	Version      string `json:"version"`
-	AuthRequired bool   `json:"auth_required"`
-	TLSRequired  bool   `json:"tls_required"`
-	MaxPayload   int64  `json:"max_payload"`
+	ID           string   `json:"server_id"`
+	Host         string   `json:"host"`
+	Port         uint     `json:"port"`
+	Version      string   `json:"version"`
+	AuthRequired bool     `json:"auth_required"`
+	TLSRequired  bool     `json:"tls_required"`
+	MaxPayload   int64    `json:"max_payload"`
+	Headers      bool     `json:"headers"`
+	ConnectURLs  []string `json:"connect_urls,omitempty"`
+	LameDuckMode bool     `json:"ldm,omitempty"`
+	CID          uint64   `json:"client_id,omitempty"`
 }
 
 type testAsyncClient struct {
@@ -69,12 +77,21 @@ func (c *testAsyncClient) parseAndClose(proto []byte) {
 }
 
 func createClientAsync(ch chan *client, s *Server, cli net.Conn) {
-	s.grWG.Add(1)
+	// Normally, those type of clients are used against non running servers.
+	// However, some don't, which would then cause the writeLoop to be
+	// started twice for the same client (since createClient() start both
+	// read and write loop if it is detected as running).
+	startWriteLoop := !s.isRunning()
+	if startWriteLoop {
+		s.grWG.Add(1)
+	}
 	go func() {
 		c := s.createClient(cli)
 		// Must be here to suppress +OK
 		c.opts.Verbose = false
-		go c.writeLoop()
+		if startWriteLoop {
+			go c.writeLoop()
+		}
 		ch <- c
 	}()
 }
@@ -118,6 +135,7 @@ func genAsyncParser(c *client) (func(string), chan bool) {
 
 var defaultServerOptions = Options{
 	Host:                  "127.0.0.1",
+	Port:                  -1,
 	Trace:                 true,
 	Debug:                 true,
 	DisableShortFirstPing: true,
@@ -153,7 +171,7 @@ func checkClientsCount(t *testing.T, s *Server, expected int) {
 
 func checkAccClientsCount(t *testing.T, acc *Account, expected int) {
 	t.Helper()
-	checkFor(t, 2*time.Second, 10*time.Millisecond, func() error {
+	checkFor(t, 4*time.Second, 10*time.Millisecond, func() error {
 		if nc := acc.NumConnections(); nc != expected {
 			return fmt.Errorf("Expected account %q to have %v clients, got %v",
 				acc.Name, expected, nc)
@@ -162,8 +180,27 @@ func checkAccClientsCount(t *testing.T, acc *Account, expected int) {
 	})
 }
 
+func TestAsyncClientWithRunningServer(t *testing.T) {
+	o := DefaultOptions()
+	s := RunServer(o)
+	defer s.Shutdown()
+
+	c, _, _ := newClientForServer(s)
+	defer c.close()
+
+	buf := make([]byte, 1000000)
+	writeLoopTxt := fmt.Sprintf("writeLoop(%p)", c.client)
+	checkFor(t, time.Second, 15*time.Millisecond, func() error {
+		n := runtime.Stack(buf, true)
+		if count := strings.Count(string(buf[:n]), writeLoopTxt); count != 1 {
+			return fmt.Errorf("writeLoop for client should have been started only once: %v", count)
+		}
+		return nil
+	})
+}
+
 func TestClientCreateAndInfo(t *testing.T) {
-	c, l := setUpClientWithResponse()
+	s, c, _, l := rawSetup(defaultServerOptions)
 	defer c.close()
 
 	if c.cid != 1 {
@@ -185,11 +222,247 @@ func TestClientCreateAndInfo(t *testing.T) {
 	// Sanity checks
 	if info.MaxPayload != MAX_PAYLOAD_SIZE ||
 		info.AuthRequired || info.TLSRequired ||
-		info.Port != DEFAULT_PORT {
+		int(info.Port) != s.opts.Port {
 		t.Fatalf("INFO inconsistent: %+v\n", info)
 	}
 }
 
+func TestClientNoResponderSupport(t *testing.T) {
+	opts := defaultServerOptions
+	s := New(&opts)
+
+	c, _, _ := newClientForServer(s)
+	defer c.close()
+
+	// Force header support if you want to do no_responders. Make sure headers are set.
+	if err := c.parse([]byte("CONNECT {\"no_responders\":true}\r\n")); err == nil {
+		t.Fatalf("Expected error")
+	}
+
+	c, cr, _ := newClientForServer(s)
+	defer c.close()
+
+	c.parseAsync("CONNECT {\"headers\":true, \"no_responders\":true}\r\nSUB reply 1\r\nPUB foo reply 2\r\nok\r\n")
+
+	l, err := cr.ReadString('\n')
+	if err != nil {
+		t.Fatalf("Error receiving msg from server: %v\n", err)
+	}
+
+	am := hmsgPat.FindAllStringSubmatch(l, -1)
+	if len(am) == 0 {
+		t.Fatalf("Did not get a match for %q", l)
+	}
+	checkPayload(cr, []byte("NATS/1.0 503\r\n\r\n"), t)
+}
+
+func TestServerHeaderSupport(t *testing.T) {
+	opts := defaultServerOptions
+	s := New(&opts)
+
+	c, _, l := newClientForServer(s)
+	defer c.close()
+
+	if !strings.HasPrefix(l, "INFO ") {
+		t.Fatalf("INFO response incorrect: %s\n", l)
+	}
+	var info serverInfo
+	if err := json.Unmarshal([]byte(l[5:]), &info); err != nil {
+		t.Fatalf("Could not parse INFO json: %v\n", err)
+	}
+	if !info.Headers {
+		t.Fatalf("Expected by default for header support to be enabled")
+	}
+
+	opts.NoHeaderSupport = true
+	opts.Port = -1
+	s = New(&opts)
+
+	c, _, l = newClientForServer(s)
+	defer c.close()
+
+	if err := json.Unmarshal([]byte(l[5:]), &info); err != nil {
+		t.Fatalf("Could not parse INFO json: %v\n", err)
+	}
+	if info.Headers {
+		t.Fatalf("Expected header support to be disabled")
+	}
+}
+
+// This test specifically is not testing how headers are encoded in a raw msg.
+// It wants to make sure the serve and clients agreement on when to use headers
+// is bi-directional and functions properly.
+func TestClientHeaderSupport(t *testing.T) {
+	opts := defaultServerOptions
+	s := New(&opts)
+
+	c, _, _ := newClientForServer(s)
+	defer c.close()
+
+	// Even though the server supports headers we need to explicitly say we do in the
+	// CONNECT. If we do not we should get an error.
+	if err := c.parse([]byte("CONNECT {}\r\nHPUB foo 0 2\r\nok\r\n")); err != ErrMsgHeadersNotSupported {
+		t.Fatalf("Expected to receive an error, got %v", err)
+	}
+
+	// This should succeed.
+	c, _, _ = newClientForServer(s)
+	defer c.close()
+
+	if err := c.parse([]byte("CONNECT {\"headers\":true}\r\nHPUB foo 0 2\r\nok\r\n")); err != nil {
+		t.Fatalf("Unexpected error %v", err)
+	}
+
+	// Now start a server without support.
+	opts.NoHeaderSupport = true
+	opts.Port = -1
+	s = New(&opts)
+
+	c, _, _ = newClientForServer(s)
+	defer c.close()
+	if err := c.parse([]byte("CONNECT {\"headers\":true}\r\nHPUB foo 0 2\r\nok\r\n")); err != ErrMsgHeadersNotSupported {
+		t.Fatalf("Expected to receive an error, got %v", err)
+	}
+}
+
+var hmsgPat = regexp.MustCompile(`HMSG\s+([^\s]+)\s+([^\s]+)\s+(([^\s]+)[^\S\r\n]+)?(\d+)[^\S\r\n]+(\d+)\r\n`)
+
+func TestClientHeaderDeliverMsg(t *testing.T) {
+	opts := defaultServerOptions
+	s := New(&opts)
+
+	c, cr, _ := newClientForServer(s)
+	defer c.close()
+
+	connect := "CONNECT {\"headers\":true}"
+	subOp := "SUB foo 1"
+	pubOp := "HPUB foo 12 14\r\nName:Derek\r\nOK\r\n"
+	cmd := strings.Join([]string{connect, subOp, pubOp}, "\r\n")
+
+	c.parseAsync(cmd)
+	l, err := cr.ReadString('\n')
+	if err != nil {
+		t.Fatalf("Error receiving msg from server: %v\n", err)
+	}
+
+	am := hmsgPat.FindAllStringSubmatch(l, -1)
+	if len(am) == 0 {
+		t.Fatalf("Did not get a match for %q", l)
+	}
+	matches := am[0]
+	if len(matches) != 7 {
+		t.Fatalf("Did not get correct # matches: %d vs %d\n", len(matches), 7)
+	}
+	if matches[SUB_INDEX] != "foo" {
+		t.Fatalf("Did not get correct subject: '%s'\n", matches[SUB_INDEX])
+	}
+	if matches[SID_INDEX] != "1" {
+		t.Fatalf("Did not get correct sid: '%s'\n", matches[SID_INDEX])
+	}
+	if matches[HDR_INDEX] != "12" {
+		t.Fatalf("Did not get correct msg length: '%s'\n", matches[HDR_INDEX])
+	}
+	if matches[TLEN_INDEX] != "14" {
+		t.Fatalf("Did not get correct msg length: '%s'\n", matches[TLEN_INDEX])
+	}
+	checkPayload(cr, []byte("Name:Derek\r\nOK\r\n"), t)
+}
+
+var smsgPat = regexp.MustCompile(`^MSG\s+([^\s]+)\s+([^\s]+)\s+(([^\s]+)[^\S\r\n]+)?(\d+)\r\n`)
+
+func TestClientHeaderDeliverStrippedMsg(t *testing.T) {
+	opts := defaultServerOptions
+	s := New(&opts)
+
+	c, _, _ := newClientForServer(s)
+	defer c.close()
+
+	b, br, _ := newClientForServer(s)
+	defer b.close()
+
+	// Does not support headers
+	b.parseAsync("SUB foo 1\r\nPING\r\n")
+	if _, err := br.ReadString('\n'); err != nil {
+		t.Fatalf("Error receiving msg from server: %v\n", err)
+	}
+
+	connect := "CONNECT {\"headers\":true}"
+	pubOp := "HPUB foo 12 14\r\nName:Derek\r\nOK\r\n"
+	cmd := strings.Join([]string{connect, pubOp}, "\r\n")
+	c.parseAsync(cmd)
+	// Read from 'b' client.
+	l, err := br.ReadString('\n')
+	if err != nil {
+		t.Fatalf("Error receiving msg from server: %v\n", err)
+	}
+	am := smsgPat.FindAllStringSubmatch(l, -1)
+	if len(am) == 0 {
+		t.Fatalf("Did not get a correct match for %q", l)
+	}
+	matches := am[0]
+	if len(matches) != 6 {
+		t.Fatalf("Did not get correct # matches: %d vs %d\n", len(matches), 6)
+	}
+	if matches[SUB_INDEX] != "foo" {
+		t.Fatalf("Did not get correct subject: '%s'\n", matches[SUB_INDEX])
+	}
+	if matches[SID_INDEX] != "1" {
+		t.Fatalf("Did not get correct sid: '%s'\n", matches[SID_INDEX])
+	}
+	if matches[LEN_INDEX] != "2" {
+		t.Fatalf("Did not get correct msg length: '%s'\n", matches[LEN_INDEX])
+	}
+	checkPayload(br, []byte("OK\r\n"), t)
+	if br.Buffered() != 0 {
+		t.Fatalf("Expected no extra bytes to be buffered, got %d", br.Buffered())
+	}
+}
+
+func TestClientHeaderDeliverQueueSubStrippedMsg(t *testing.T) {
+	opts := defaultServerOptions
+	s := New(&opts)
+
+	c, _, _ := newClientForServer(s)
+	defer c.close()
+
+	b, br, _ := newClientForServer(s)
+	defer b.close()
+
+	// Does not support headers
+	b.parseAsync("SUB foo bar 1\r\nPING\r\n")
+	if _, err := br.ReadString('\n'); err != nil {
+		t.Fatalf("Error receiving msg from server: %v\n", err)
+	}
+
+	connect := "CONNECT {\"headers\":true}"
+	pubOp := "HPUB foo 12 14\r\nName:Derek\r\nOK\r\n"
+	cmd := strings.Join([]string{connect, pubOp}, "\r\n")
+	c.parseAsync(cmd)
+	// Read from 'b' client.
+	l, err := br.ReadString('\n')
+	if err != nil {
+		t.Fatalf("Error receiving msg from server: %v\n", err)
+	}
+	am := smsgPat.FindAllStringSubmatch(l, -1)
+	if len(am) == 0 {
+		t.Fatalf("Did not get a correct match for %q", l)
+	}
+	matches := am[0]
+	if len(matches) != 6 {
+		t.Fatalf("Did not get correct # matches: %d vs %d\n", len(matches), 6)
+	}
+	if matches[SUB_INDEX] != "foo" {
+		t.Fatalf("Did not get correct subject: '%s'\n", matches[SUB_INDEX])
+	}
+	if matches[SID_INDEX] != "1" {
+		t.Fatalf("Did not get correct sid: '%s'\n", matches[SID_INDEX])
+	}
+	if matches[LEN_INDEX] != "2" {
+		t.Fatalf("Did not get correct msg length: '%s'\n", matches[LEN_INDEX])
+	}
+	checkPayload(br, []byte("OK\r\n"), t)
+}
+
 func TestNonTLSConnectionState(t *testing.T) {
 	_, c, _ := setupClient()
 	defer c.close()
@@ -212,7 +485,7 @@ func TestClientConnect(t *testing.T) {
 	if c.state != OP_START {
 		t.Fatalf("Expected state of OP_START vs %d\n", c.state)
 	}
-	if !reflect.DeepEqual(c.opts, clientOpts{Verbose: true, Pedantic: true, Echo: false}) {
+	if !reflect.DeepEqual(c.opts, ClientOpts{Verbose: true, Pedantic: true, Echo: false}) {
 		t.Fatalf("Did not parse connect options correctly: %+v\n", c.opts)
 	}
 
@@ -226,7 +499,7 @@ func TestClientConnect(t *testing.T) {
 	if c.state != OP_START {
 		t.Fatalf("Expected state of OP_START vs %d\n", c.state)
 	}
-	if !reflect.DeepEqual(c.opts, clientOpts{Echo: true, Verbose: true, Pedantic: true, Username: "derek", Password: "foo"}) {
+	if !reflect.DeepEqual(c.opts, ClientOpts{Echo: true, Verbose: true, Pedantic: true, Username: "derek", Password: "foo"}) {
 		t.Fatalf("Did not parse connect options correctly: %+v\n", c.opts)
 	}
 
@@ -241,7 +514,7 @@ func TestClientConnect(t *testing.T) {
 		t.Fatalf("Expected state of OP_START vs %d\n", c.state)
 	}
 
-	if !reflect.DeepEqual(c.opts, clientOpts{Echo: true, Verbose: true, Pedantic: true, Username: "derek", Password: "foo", Name: "router"}) {
+	if !reflect.DeepEqual(c.opts, ClientOpts{Echo: true, Verbose: true, Pedantic: true, Username: "derek", Password: "foo", Name: "router"}) {
 		t.Fatalf("Did not parse connect options correctly: %+v\n", c.opts)
 	}
 
@@ -256,7 +529,7 @@ func TestClientConnect(t *testing.T) {
 		t.Fatalf("Expected state of OP_START vs %d\n", c.state)
 	}
 
-	if !reflect.DeepEqual(c.opts, clientOpts{Echo: true, Verbose: true, Pedantic: true, Authorization: "YZZ222", Name: "router"}) {
+	if !reflect.DeepEqual(c.opts, ClientOpts{Echo: true, Verbose: true, Pedantic: true, Token: "YZZ222", Name: "router"}) {
 		t.Fatalf("Did not parse connect options correctly: %+v\n", c.opts)
 	}
 }
@@ -274,7 +547,7 @@ func TestClientConnectProto(t *testing.T) {
 	if c.state != OP_START {
 		t.Fatalf("Expected state of OP_START vs %d\n", c.state)
 	}
-	if !reflect.DeepEqual(c.opts, clientOpts{Echo: true, Verbose: true, Pedantic: true, Protocol: ClientProtoZero}) {
+	if !reflect.DeepEqual(c.opts, ClientOpts{Echo: true, Verbose: true, Pedantic: true, Protocol: ClientProtoZero}) {
 		t.Fatalf("Did not parse connect options correctly: %+v\n", c.opts)
 	}
 
@@ -287,7 +560,7 @@ func TestClientConnectProto(t *testing.T) {
 	if c.state != OP_START {
 		t.Fatalf("Expected state of OP_START vs %d\n", c.state)
 	}
-	if !reflect.DeepEqual(c.opts, clientOpts{Echo: true, Verbose: true, Pedantic: true, Protocol: ClientProtoInfo}) {
+	if !reflect.DeepEqual(c.opts, ClientOpts{Echo: true, Verbose: true, Pedantic: true, Protocol: ClientProtoInfo}) {
 		t.Fatalf("Did not parse connect options correctly: %+v\n", c.opts)
 	}
 	if c.opts.Protocol != ClientProtoInfo {
@@ -363,6 +636,8 @@ const (
 	SID_INDEX   = 2
 	REPLY_INDEX = 4
 	LEN_INDEX   = 5
+	HDR_INDEX   = 5
+	TLEN_INDEX  = 6
 )
 
 func grabPayload(cr *bufio.Reader, expected int) []byte {
@@ -573,6 +848,54 @@ func TestSplitSubjectQueue(t *testing.T) {
 	}
 }
 
+func TestTypeString(t *testing.T) {
+	cases := []struct {
+		intType    int
+		stringType string
+	}{
+		{
+			intType:    CLIENT,
+			stringType: "Client",
+		},
+		{
+			intType:    ROUTER,
+			stringType: "Router",
+		},
+		{
+			intType:    GATEWAY,
+			stringType: "Gateway",
+		},
+		{
+			intType:    LEAF,
+			stringType: "Leafnode",
+		},
+		{
+			intType:    JETSTREAM,
+			stringType: "JetStream",
+		},
+		{
+			intType:    ACCOUNT,
+			stringType: "Account",
+		},
+		{
+			intType:    SYSTEM,
+			stringType: "System",
+		},
+		{
+			intType:    -1,
+			stringType: "Unknown Type",
+		},
+	}
+	for _, cs := range cases {
+		c := &client{kind: cs.intType}
+		typeStringVal := c.kindString()
+
+		if typeStringVal != cs.stringType {
+			t.Fatalf("Expected typeString  value  %q, but instead received %q", cs.stringType, typeStringVal)
+		}
+	}
+}
+
 func TestQueueSubscribePermissions(t *testing.T) {
 	cases := []struct {
 		name    string
@@ -1307,65 +1630,6 @@ func TestDynamicBuffers(t *testing.T) {
 	checkBuffers(minBufSize, minBufSize)
 }
 
-// Similar to the routed version. Make sure we receive all of the
-// messages with auto-unsubscribe enabled.
-func TestQueueAutoUnsubscribe(t *testing.T) {
-	opts := DefaultOptions()
-	s := RunServer(opts)
-	defer s.Shutdown()
-
-	nc, err := nats.Connect(fmt.Sprintf("nats://%s:%d", opts.Host, opts.Port))
-	if err != nil {
-		t.Fatalf("Error on connect: %v", err)
-	}
-	defer nc.Close()
-
-	rbar := int32(0)
-	barCb := func(m *nats.Msg) {
-		atomic.AddInt32(&rbar, 1)
-	}
-	rbaz := int32(0)
-	bazCb := func(m *nats.Msg) {
-		atomic.AddInt32(&rbaz, 1)
-	}
-
-	// Create 1000 subscriptions with auto-unsubscribe of 1.
-	// Do two groups, one bar and one baz.
-	for i := 0; i < 1000; i++ {
-		qsub, err := nc.QueueSubscribe("foo", "bar", barCb)
-		if err != nil {
-			t.Fatalf("Error on subscribe: %v", err)
-		}
-		if err := qsub.AutoUnsubscribe(1); err != nil {
-			t.Fatalf("Error on auto-unsubscribe: %v", err)
-		}
-		qsub, err = nc.QueueSubscribe("foo", "baz", bazCb)
-		if err != nil {
-			t.Fatalf("Error on subscribe: %v", err)
-		}
-		if err := qsub.AutoUnsubscribe(1); err != nil {
-			t.Fatalf("Error on auto-unsubscribe: %v", err)
-		}
-	}
-	nc.Flush()
-
-	expected := int32(1000)
-	for i := int32(0); i < expected; i++ {
-		nc.Publish("foo", []byte("Don't Drop Me!"))
-	}
-	nc.Flush()
-
-	checkFor(t, 5*time.Second, 10*time.Millisecond, func() error {
-		nbar := atomic.LoadInt32(&rbar)
-		nbaz := atomic.LoadInt32(&rbaz)
-		if nbar == expected && nbaz == expected {
-			return nil
-		}
-		return fmt.Errorf("Did not receive all %d queue messages, received %d for 'bar' and %d for 'baz'",
-			expected, atomic.LoadInt32(&rbar), atomic.LoadInt32(&rbaz))
-	})
-}
-
 func TestClientTraceRace(t *testing.T) {
 	opts := DefaultOptions()
 	s := RunServer(opts)
@@ -1417,7 +1681,7 @@ func TestClientUserInfo(t *testing.T) {
 	pnkey := "UD6AYQSOIN2IN5OGC6VQZCR4H3UFMIOXSW6NNS6N53CLJA4PB56CEJJI"
 	c := &client{
 		cid: 1024,
-		opts: clientOpts{
+		opts: ClientOpts{
 			Nkey: pnkey,
 		},
 	}
@@ -1429,7 +1693,7 @@ func TestClientUserInfo(t *testing.T) {
 
 	c = &client{
 		cid: 1024,
-		opts: clientOpts{
+		opts: ClientOpts{
 			Username: "foo",
 		},
 	}
@@ -1441,7 +1705,7 @@ func TestClientUserInfo(t *testing.T) {
 
 	c = &client{
 		cid:  1024,
-		opts: clientOpts{},
+		opts: ClientOpts{},
 	}
 	got = c.getAuthUser()
 	expected = `User "N/A"`
@@ -1462,15 +1726,6 @@ func (l *captureWarnLogger) Warnf(format string, v ...interface{}) {
 	}
 }
 
-type pauseWriteConn struct {
-	net.Conn
-}
-
-func (swc *pauseWriteConn) Write(b []byte) (int, error) {
-	time.Sleep(250 * time.Millisecond)
-	return swc.Conn.Write(b)
-}
-
 func TestReadloopWarning(t *testing.T) {
 	readLoopReportThreshold = 100 * time.Millisecond
 	defer func() { readLoopReportThreshold = readLoopReport }()
@@ -1492,10 +1747,15 @@ func TestReadloopWarning(t *testing.T) {
 	sender := natsConnect(t, url)
 	defer sender.Close()
 
+	wg := sync.WaitGroup{}
+	wg.Add(1)
 	c := s.getClient(cid)
 	c.mu.Lock()
-	c.nc = &pauseWriteConn{Conn: c.nc}
-	c.mu.Unlock()
+	go func() {
+		defer wg.Done()
+		time.Sleep(250 * time.Millisecond)
+		c.mu.Unlock()
+	}()
 
 	natsPub(t, sender, "foo", make([]byte, 100))
 	natsFlush(t, sender)
@@ -1508,6 +1768,7 @@ func TestReadloopWarning(t *testing.T) {
 	case <-time.After(2 * time.Second):
 		t.Fatalf("No warning printed")
 	}
+	wg.Wait()
 }
 
 func TestTraceMsg(t *testing.T) {
@@ -1555,7 +1816,7 @@ func TestTraceMsg(t *testing.T) {
 
 		c.traceMsg(ut.Msg)
 
-		got := c.srv.logging.logger.(*DummyLogger).msg
+		got := c.srv.logging.logger.(*DummyLogger).Msg
 		if !reflect.DeepEqual(ut.Wanted, got) {
 			t.Errorf("Desc: %s. Msg %q. Traced msg want: %s, got: %s", ut.Desc, ut.Msg, ut.Wanted, got)
 		}
@@ -1780,19 +2041,27 @@ func TestNoClientLeakOnSlowConsumer(t *testing.T) {
 	}
 	defer c.Close()
 
-	var buf [1024]byte
+	cr := bufio.NewReader(c)
+
 	// Wait for INFO...
-	c.Read(buf[:])
+	line, _, _ := cr.ReadLine()
+	var info serverInfo
+	if err = json.Unmarshal(line[5:], &info); err != nil {
+		t.Fatalf("Could not parse INFO json: %v\n", err)
+	}
 
 	// Send our connect
-	if _, err := c.Write([]byte("CONNECT {}\r\nSUB foo 1\r\nPING\r\n")); err != nil {
+	if _, err := c.Write([]byte("CONNECT {\"verbose\": false}\r\nSUB foo 1\r\nPING\r\n")); err != nil {
 		t.Fatalf("Error sending CONNECT and SUB: %v", err)
 	}
 	// Wait for PONG
-	c.Read(buf[:])
+	line, _, _ = cr.ReadLine()
+	if string(line) != "PONG" {
+		t.Fatalf("Expected 'PONG' but got %q", line)
+	}
 
 	// Get the client from server map
-	cli := s.GetClient(1)
+	cli := s.GetClient(info.CID)
 	if cli == nil {
 		t.Fatalf("No client registered")
 	}
@@ -1909,9 +2178,10 @@ func TestClientIPv6Address(t *testing.T) {
 	}
 	defer nc.Close()
 
-	c := s.GetClient(1)
+	cid, _ := nc.GetClientID()
+	c := s.GetClient(cid)
 	c.mu.Lock()
-	ncs := c.ncs
+	ncs := c.String()
 	c.mu.Unlock()
 	if !strings.HasPrefix(ncs, "[::1]") {
 		t.Fatalf("Wrong string representation of an IPv6 address: %q", ncs)
@@ -2045,34 +2315,55 @@ func TestCloseConnectionLogsReason(t *testing.T) {
 }
 
 func TestCloseConnectionVeryEarly(t *testing.T) {
-	o := DefaultOptions()
-	s := RunServer(o)
-	defer s.Shutdown()
+	for _, test := range []struct {
+		name   string
+		useTLS bool
+	}{
+		{"no_tls", false},
+		{"tls", true},
+	} {
+		t.Run(test.name, func(t *testing.T) {
+			o := DefaultOptions()
+			if test.useTLS {
+				tc := &TLSConfigOpts{
+					CertFile: "../test/configs/certs/server-cert.pem",
+					KeyFile:  "../test/configs/certs/server-key.pem",
+					CaFile:   "../test/configs/certs/ca.pem",
+				}
+				tlsConfig, err := GenTLSConfig(tc)
+				if err != nil {
+					t.Fatalf("Error generating tls config: %v", err)
+				}
+				o.TLSConfig = tlsConfig
+			}
+			s := RunServer(o)
+			defer s.Shutdown()
 
-	// The issue was with a connection that would break right when
-	// server was sending the INFO. Creating a bare TCP connection
-	// and closing it right away won't help reproduce the problem.
-	// So testing in 2 steps.
+			// The issue was with a connection that would break right when
+			// server was sending the INFO. Creating a bare TCP connection
+			// and closing it right away won't help reproduce the problem.
+			// So testing in 2 steps.
 
-	// Get a normal TCP connection to the server.
-	c, err := net.Dial("tcp", fmt.Sprintf("127.0.0.1:%d", o.Port))
-	if err != nil {
-		s.mu.Unlock()
-		t.Fatalf("Unable to create tcp connection")
+			// Get a normal TCP connection to the server.
+			c, err := net.Dial("tcp", fmt.Sprintf("127.0.0.1:%d", o.Port))
+			if err != nil {
+				t.Fatalf("Unable to create tcp connection")
+			}
+			// Now close it.
+			c.Close()
+
+			// Wait that num clients falls to 0.
+			checkClientsCount(t, s, 0)
+
+			// Call again with this closed connection. Alternatively, we
+			// would have to call with a fake connection that implements
+			// net.Conn but returns an error on Write.
+			s.createClient(c)
+
+			// This connection should not have been added to the server.
+			checkClientsCount(t, s, 0)
+		})
 	}
-	// Now close it.
-	c.Close()
-
-	// Wait that num clients falls to 0.
-	checkClientsCount(t, s, 0)
-
-	// Call again with this closed connection. Alternatively, we
-	// would have to call with a fake connection that implements
-	// net.Conn but returns an error on Write.
-	s.createClient(c)
-
-	// This connection should not have been added to the server.
-	checkClientsCount(t, s, 0)
 }
 
 type connAddrString struct {
@@ -2103,14 +2394,24 @@ func TestClientConnectionName(t *testing.T) {
 		name    string
 		kind    int
 		kindStr string
+		ws      bool
+		mqtt    bool
 	}{
-		{"client", CLIENT, "cid:"},
-		{"route", ROUTER, "rid:"},
-		{"gateway", GATEWAY, "gid:"},
-		{"leafnode", LEAF, "lid:"},
+		{"client", CLIENT, "cid:", false, false},
+		{"ws client", CLIENT, "wid:", true, false},
+		{"mqtt client", CLIENT, "mid:", false, true},
+		{"route", ROUTER, "rid:", false, false},
+		{"gateway", GATEWAY, "gid:", false, false},
+		{"leafnode", LEAF, "lid:", false, false},
 	} {
 		t.Run(test.name, func(t *testing.T) {
 			c := &client{srv: s, nc: &connString{}, kind: test.kind}
+			if test.ws {
+				c.ws = &websocket{}
+			}
+			if test.mqtt {
+				c.mqtt = &mqtt{}
+			}
 			c.initClient()
 
 			if host := "fe80::abc:def:ghi:123%utun0"; host != c.host {
@@ -2123,15 +2424,15 @@ func TestClientConnectionName(t *testing.T) {
 			checkLog := func(suffix string) {
 				t.Helper()
 				l.Lock()
-				msg := l.msg
+				msg := l.Msg
 				l.Unlock()
 				if strings.Contains(msg, "(MISSING)") {
 					t.Fatalf("conn name was not escaped properly, got MISSING: %s", msg)
 				}
-				if !strings.Contains(l.msg, test.kindStr) {
+				if !strings.Contains(l.Msg, test.kindStr) {
 					t.Fatalf("expected kind to be %q, got: %s", test.kindStr, msg)
 				}
-				if !strings.HasSuffix(l.msg, suffix) {
+				if !strings.HasSuffix(l.Msg, suffix) {
 					t.Fatalf("expected statement to end with %q, got %s", suffix, msg)
 				}
 			}
@@ -2147,3 +2448,159 @@ func TestClientConnectionName(t *testing.T) {
 		})
 	}
 }
+
+func TestClientLimits(t *testing.T) {
+	accKp, err := nkeys.CreateAccount()
+	if err != nil {
+		t.Fatalf("Error creating account key: %v", err)
+	}
+	uKp, err := nkeys.CreateUser()
+	if err != nil {
+		t.Fatalf("Error creating user key: %v", err)
+	}
+	uPub, err := uKp.PublicKey()
+	if err != nil {
+		t.Fatalf("Error obtaining publicKey: %v", err)
+	}
+	s, err := NewServer(DefaultOptions())
+	if err != nil {
+		t.Fatalf("Error creating server: %v", err)
+	}
+	for _, test := range []struct {
+		client int32
+		acc    int32
+		srv    int32
+		expect int32
+	}{
+		// all identical
+		{1, 1, 1, 1},
+		{-1, -1, 0, -1},
+		// only one value unlimited
+		{1, -1, 0, 1},
+		{-1, 1, 0, 1},
+		{-1, -1, 1, 1},
+		// all combinations of distinct values
+		{1, 2, 3, 1},
+		{1, 3, 2, 1},
+		{2, 1, 3, 1},
+		{2, 3, 1, 1},
+		{3, 1, 2, 1},
+		{3, 2, 1, 1},
+	} {
+		t.Run("", func(t *testing.T) {
+			s.opts.MaxPayload = test.srv
+			s.opts.MaxSubs = int(test.srv)
+			c := &client{srv: s, acc: &Account{
+				limits: limits{mpay: test.acc, msubs: test.acc},
+			}}
+			uc := jwt.NewUserClaims(uPub)
+			uc.Limits.Subs = int64(test.client)
+			uc.Limits.Payload = int64(test.client)
+			c.opts.JWT, err = uc.Encode(accKp)
+			if err != nil {
+				t.Fatalf("Error encoding jwt: %v", err)
+			}
+			c.applyAccountLimits()
+			if c.mpay != test.expect {
+				t.Fatalf("payload %d not as ecpected %d", c.mpay, test.expect)
+			}
+			if c.msubs != test.expect {
+				t.Fatalf("subscriber %d not as ecpected %d", c.msubs, test.expect)
+			}
+		})
+	}
+}
+
+func TestClientClampMaxSubsErrReport(t *testing.T) {
+	maxSubLimitReportThreshold = int64(100 * time.Millisecond)
+	defer func() { maxSubLimitReportThreshold = defaultMaxSubLimitReportThreshold }()
+
+	o1 := DefaultOptions()
+	o1.MaxSubs = 1
+	o1.LeafNode.Host = "127.0.0.1"
+	o1.LeafNode.Port = -1
+	s1 := RunServer(o1)
+	defer s1.Shutdown()
+
+	l := &captureErrorLogger{errCh: make(chan string, 10)}
+	s1.SetLogger(l, false, false)
+
+	o2 := DefaultOptions()
+	o2.Cluster.Name = "xyz"
+	u, _ := url.Parse(fmt.Sprintf("nats://127.0.0.1:%d", o1.LeafNode.Port))
+	o2.LeafNode.Remotes = []*RemoteLeafOpts{{URLs: []*url.URL{u}}}
+	s2 := RunServer(o2)
+	defer s2.Shutdown()
+
+	checkLeafNodeConnected(t, s1)
+	checkLeafNodeConnected(t, s2)
+
+	nc := natsConnect(t, s2.ClientURL())
+	defer nc.Close()
+	natsSubSync(t, nc, "foo")
+	natsSubSync(t, nc, "bar")
+
+	// Make sure we receive only 1
+	check := func() {
+		t.Helper()
+		for i := 0; i < 2; i++ {
+			select {
+			case errStr := <-l.errCh:
+				if i > 0 {
+					t.Fatalf("Should not have logged a second time: %s", errStr)
+				}
+				if !strings.Contains(errStr, "maximum subscriptions") {
+					t.Fatalf("Unexpected error: %s", errStr)
+				}
+			case <-time.After(300 * time.Millisecond):
+				if i == 0 {
+					t.Fatal("Error should have been logged")
+				}
+			}
+		}
+	}
+	check()
+
+	// The above will have waited long enough to clear the report threshold.
+	// So create two new subs and check again that we get only 1 report.
+	natsSubSync(t, nc, "baz")
+	natsSubSync(t, nc, "bat")
+	check()
+}
+
+func TestClientDenySysGroupSub(t *testing.T) {
+	s := RunServer(DefaultOptions())
+	defer s.Shutdown()
+
+	nc, err := nats.Connect(s.ClientURL(), nats.ErrorHandler(func(*nats.Conn, *nats.Subscription, error) {}))
+	require_NoError(t, err)
+	defer nc.Close()
+
+	_, err = nc.QueueSubscribeSync("foo", sysGroup)
+	require_NoError(t, err)
+	nc.Flush()
+	err = nc.LastError()
+	require_Error(t, err)
+	require_Contains(t, err.Error(), "Permissions Violation")
+}
+
+func TestClientAuthRequiredNoAuthUser(t *testing.T) {
+	conf := createConfFile(t, []byte(`
+		listen: 127.0.0.1:-1
+		accounts: {
+			A: { users: [ { user: user, password: pass } ] }
+		}
+		no_auth_user: user
+	`))
+
+	s, _ := RunServerWithConfig(conf)
+	defer s.Shutdown()
+
+	nc, err := nats.Connect(s.ClientURL())
+	require_NoError(t, err)
+	defer nc.Close()
+
+	if nc.AuthRequired() {
+		t.Fatalf("Expected AuthRequired to be false due to 'no_auth_user'")
+	}
+}
