@@ -19,9 +19,7 @@ import (
 	"encoding/binary"
 	"encoding/hex"
 	"encoding/json"
-	"errors"
 	"fmt"
-	"io/ioutil"
 	"math"
 	"os"
 	"path/filepath"
@@ -35,16 +33,16 @@ import (
 	"github.com/nats-io/nats-server/v2/server/sysmem"
 	"github.com/nats-io/nkeys"
 	"github.com/nats-io/nuid"
-	"golang.org/x/crypto/chacha20poly1305"
 )
 
 // JetStreamConfig determines this server's configuration.
 // MaxMemory and MaxStore are in bytes.
 type JetStreamConfig struct {
-	MaxMemory int64  `json:"max_memory"`
-	MaxStore  int64  `json:"max_storage"`
-	StoreDir  string `json:"store_dir,omitempty"`
-	Domain    string `json:"domain,omitempty"`
+	MaxMemory  int64  `json:"max_memory"`
+	MaxStore   int64  `json:"max_storage"`
+	StoreDir   string `json:"store_dir,omitempty"`
+	Domain     string `json:"domain,omitempty"`
+	CompressOK bool   `json:"compress_ok,omitempty"`
 }
 
 // Statistics about JetStream for this server.
@@ -101,15 +99,21 @@ type jetStream struct {
 	storeReserved int64
 	memUsed       int64
 	storeUsed     int64
+	clustered     int32
 	mu            sync.RWMutex
 	srv           *Server
 	config        JetStreamConfig
 	cluster       *jetStreamCluster
 	accounts      map[string]*jsAccount
 	apiSubs       *Sublist
-	standAlone    bool
-	disabled      bool
-	oos           bool
+	started       time.Time
+
+	// System level request to purge a stream move
+	accountPurge   *subscription
+	metaRecovering bool
+	standAlone     bool
+	disabled       bool
+	oos            bool
 }
 
 type remoteUsage struct {
@@ -132,25 +136,27 @@ type jsAccount struct {
 	mu        sync.RWMutex
 	js        *jetStream
 	account   *Account
-	limits    map[string]JetStreamAccountLimits // indexed by tierName
-	usage     map[string]*jsaStorage            // indexed by tierName
-	apiTotal  uint64
-	apiErrors uint64
-	usageApi  uint64
-	usageErr  uint64
-	rusage    map[string]*remoteUsage // indexed by node id
 	storeDir  string
 	streams   map[string]*stream
 	templates map[string]*streamTemplate
 	store     TemplateStore
 
-	// Cluster support
+	// From server
+	sendq *ipQueue // of *pubMsg
+
+	// Usage/limits related fields that will be protected by usageMu
+	usageMu    sync.RWMutex
+	limits     map[string]JetStreamAccountLimits // indexed by tierName
+	usage      map[string]*jsaStorage            // indexed by tierName
+	rusage     map[string]*remoteUsage           // indexed by node id
+	apiTotal   uint64
+	apiErrors  uint64
+	usageApi   uint64
+	usageErr   uint64
 	updatesPub string
 	updatesSub *subscription
-	// From server
-	sendq   *ipQueue // of *pubMsg
-	lupdate time.Time
-	utimer  *time.Timer
+	lupdate    time.Time
+	utimer     *time.Timer
 }
 
 // Track general usage for this account.
@@ -197,10 +203,6 @@ func (s *Server) EnableJetStream(config *JetStreamConfig) error {
 		return err
 	}
 
-	if ek := s.getOpts().JetStreamKey; ek != _EMPTY_ {
-		s.Warnf("JetStream Encryption is Beta")
-	}
-
 	return s.enableJetStream(cfg)
 }
 
@@ -226,9 +228,9 @@ func (s *Server) jsKeyGen(info string) keyGen {
 }
 
 // Decode the encrypted metafile.
-func (s *Server) decryptMeta(ekey, buf []byte, acc, context string) ([]byte, error) {
-	if len(ekey) != metaKeySize {
-		return nil, errors.New("bad encryption key")
+func (s *Server) decryptMeta(sc StoreCipher, ekey, buf []byte, acc, context string) ([]byte, error) {
+	if len(ekey) < minMetaKeySize {
+		return nil, errBadKeySize
 	}
 	prf := s.jsKeyGen(acc)
 	if prf == nil {
@@ -238,7 +240,8 @@ func (s *Server) decryptMeta(ekey, buf []byte, acc, context string) ([]byte, err
 	if err != nil {
 		return nil, err
 	}
-	kek, err := chacha20poly1305.NewX(rb)
+
+	kek, err := genEncryptionKey(sc, rb)
 	if err != nil {
 		return nil, err
 	}
@@ -247,11 +250,10 @@ func (s *Server) decryptMeta(ekey, buf []byte, acc, context string) ([]byte, err
 	if err != nil {
 		return nil, err
 	}
-	aek, err := chacha20poly1305.NewX(seed[:])
+	aek, err := genEncryptionKey(sc, seed)
 	if err != nil {
 		return nil, err
 	}
-	ns = kek.NonceSize()
 	plain, err := aek.Open(nil, buf[:ns], buf[ns:], nil)
 	if err != nil {
 		return nil, err
@@ -262,14 +264,14 @@ func (s *Server) decryptMeta(ekey, buf []byte, acc, context string) ([]byte, err
 // Check to make sure directory has the jetstream directory.
 // We will have it properly configured here now regardless, so need to look inside.
 func (s *Server) checkStoreDir(cfg *JetStreamConfig) error {
-	fis, _ := ioutil.ReadDir(cfg.StoreDir)
+	fis, _ := os.ReadDir(cfg.StoreDir)
 	// If we have nothing underneath us, could be just starting new, but if we see this we can check.
 	if len(fis) != 0 {
 		return nil
 	}
 	// Let's check the directory above. If it has us 'jetstream' but also other stuff that we can
 	// identify as accounts then we can fix.
-	fis, _ = ioutil.ReadDir(filepath.Dir(cfg.StoreDir))
+	fis, _ = os.ReadDir(filepath.Dir(cfg.StoreDir))
 	// If just one that is us 'jetstream' and all is ok.
 	if len(fis) == 1 {
 		return nil
@@ -323,6 +325,11 @@ func (s *Server) checkStoreDir(cfg *JetStreamConfig) error {
 // enableJetStream will start up the JetStream subsystem.
 func (s *Server) enableJetStream(cfg JetStreamConfig) error {
 	js := &jetStream{srv: s, config: cfg, accounts: make(map[string]*jsAccount), apiSubs: NewSublistNoCache()}
+	s.gcbMu.Lock()
+	if s.gcbOutMax = s.getOpts().JetStreamMaxCatchup; s.gcbOutMax == 0 {
+		s.gcbOutMax = defaultMaxTotalCatchupOutBytes
+	}
+	s.gcbMu.Unlock()
 
 	s.mu.Lock()
 	s.js = js
@@ -338,7 +345,7 @@ func (s *Server) enableJetStream(cfg JetStreamConfig) error {
 		if stat == nil || !stat.IsDir() {
 			return fmt.Errorf("storage directory is not a directory")
 		}
-		tmpfile, err := ioutil.TempFile(cfg.StoreDir, "_test_")
+		tmpfile, err := os.CreateTemp(cfg.StoreDir, "_test_")
 		if err != nil {
 			return fmt.Errorf("storage directory is not writable")
 		}
@@ -366,6 +373,10 @@ func (s *Server) enableJetStream(cfg JetStreamConfig) error {
 	if cfg.Domain != _EMPTY_ {
 		s.Noticef("  Domain:          %s", cfg.Domain)
 	}
+	opts := s.getOpts()
+	if ek := opts.JetStreamKey; ek != _EMPTY_ {
+		s.Noticef("  Encryption:      %s", opts.JetStreamCipher)
+	}
 	s.Noticef("-------------------------------------------")
 
 	// Setup our internal subscriptions.
@@ -400,6 +411,9 @@ func (s *Server) enableJetStream(cfg JetStreamConfig) error {
 		}
 	}
 
+	// Mark when we are up and running.
+	js.setStarted()
+
 	return nil
 }
 
@@ -452,9 +466,13 @@ func (s *Server) restartJetStream() error {
 // checkStreamExports will check if we have the JS exports setup
 // on the system account, and if not go ahead and set them up.
 func (s *Server) checkJetStreamExports() {
-	sacc := s.SystemAccount()
-	if sacc != nil && sacc.getServiceExport(jsAllAPI) == nil {
-		s.setupJetStreamExports()
+	if sacc := s.SystemAccount(); sacc != nil {
+		sacc.mu.RLock()
+		se := sacc.getServiceExport(jsAllAPI)
+		sacc.mu.RUnlock()
+		if se == nil {
+			s.setupJetStreamExports()
+		}
 	}
 }
 
@@ -632,17 +650,20 @@ func (s *Server) configJetStream(acc *Account) error {
 	if acc == nil {
 		return nil
 	}
-	if acc.jsLimits != nil {
+	acc.mu.RLock()
+	jsLimits := acc.jsLimits
+	acc.mu.RUnlock()
+	if jsLimits != nil {
 		// Check if already enabled. This can be during a reload.
 		if acc.JetStreamEnabled() {
 			if err := acc.enableAllJetStreamServiceImportsAndMappings(); err != nil {
 				return err
 			}
-			if err := acc.UpdateJetStreamLimits(acc.jsLimits); err != nil {
+			if err := acc.UpdateJetStreamLimits(jsLimits); err != nil {
 				return err
 			}
 		} else {
-			if err := acc.EnableJetStream(acc.jsLimits); err != nil {
+			if err := acc.EnableJetStream(jsLimits); err != nil {
 				return err
 			}
 			if s.gateway.enabled {
@@ -678,6 +699,21 @@ func (s *Server) configAllJetStreamAccounts() error {
 		return nil
 	}
 
+	if s.sys != nil {
+		// clustered stream removal will perform this cleanup as well
+		// this is mainly for initial cleanup
+		saccName := s.sys.account.Name
+		accStoreDirs, _ := os.ReadDir(js.config.StoreDir)
+		for _, acc := range accStoreDirs {
+			if accName := acc.Name(); accName != saccName {
+				// no op if not empty
+				accDir := filepath.Join(js.config.StoreDir, accName)
+				os.Remove(filepath.Join(accDir, streamsDir))
+				os.Remove(accDir)
+			}
+		}
+	}
+
 	var jsAccounts []*Account
 	s.accounts.Range(func(k, v interface{}) bool {
 		jsAccounts = append(jsAccounts, v.(*Account))
@@ -696,7 +732,7 @@ func (s *Server) configAllJetStreamAccounts() error {
 
 	// Now walk all the storage we have and resolve any accounts that we did not process already.
 	// This is important in resolver/operator models.
-	fis, _ := ioutil.ReadDir(js.config.StoreDir)
+	fis, _ := os.ReadDir(js.config.StoreDir)
 	for _, fi := range fis {
 		if accName := fi.Name(); accName != _EMPTY_ {
 			// Only load up ones not already loaded since they are processed above.
@@ -713,6 +749,13 @@ func (s *Server) configAllJetStreamAccounts() error {
 	return nil
 }
 
+// Mark our started time.
+func (js *jetStream) setStarted() {
+	js.mu.Lock()
+	defer js.mu.Unlock()
+	js.started = time.Now()
+}
+
 func (js *jetStream) isEnabled() bool {
 	if js == nil {
 		return false
@@ -730,6 +773,12 @@ func (js *jetStream) setJetStreamStandAlone(isStandAlone bool) {
 	js.mu.Lock()
 	defer js.mu.Unlock()
 	js.standAlone = isStandAlone
+
+	if isStandAlone {
+		js.accountPurge, _ = js.srv.systemSubscribe(JSApiAccountPurge, _EMPTY_, false, nil, js.srv.jsLeaderAccountPurgeRequest)
+	} else if js.accountPurge != nil {
+		js.srv.sysUnsubscribe(js.accountPurge)
+	}
 }
 
 // JetStreamEnabled reports if jetstream is enabled for this server.
@@ -762,78 +811,42 @@ func (s *Server) JetStreamEnabledForDomain() bool {
 	return jsFound
 }
 
-// Will migrate off ephemerals if possible.
-// This means parent stream needs to be replicated.
-func (s *Server) migrateEphemerals() {
-	js, cc := s.getJetStreamCluster()
-	// Make sure JetStream is enabled and we are clustered.
-	if js == nil || cc == nil {
+// Will signal that all pull requests for consumers on this server are now invalid.
+func (s *Server) signalPullConsumers() {
+	js := s.getJetStream()
+	if js == nil {
 		return
 	}
 
-	var consumers []*consumerAssignment
-
-	js.mu.Lock()
-	if cc.meta == nil {
-		js.mu.Unlock()
-		return
-	}
-	ourID := cc.meta.ID()
-	for _, asa := range cc.streams {
-		for _, sa := range asa {
-			if rg := sa.Group; rg != nil && len(rg.Peers) > 1 && rg.isMember(ourID) && len(sa.consumers) > 0 {
-				for _, ca := range sa.consumers {
-					if ca.Group != nil && len(ca.Group.Peers) == 1 && ca.Group.isMember(ourID) {
-						// Need to select possible new peer from parent stream.
-						for _, p := range rg.Peers {
-							if p != ourID {
-								ca.Group.Peers = []string{p}
-								ca.Group.Preferred = p
-								consumers = append(consumers, ca)
-								break
-							}
-						}
-					}
-				}
-			}
-		}
-	}
-	js.mu.Unlock()
+	js.mu.RLock()
+	defer js.mu.RUnlock()
 
 	// In case we have stale pending requests.
-	hdr := []byte("NATS/1.0 409 Consumer Migration\r\n\r\n")
+	const hdr = "NATS/1.0 409 Server Shutdown\r\n" + JSPullRequestPendingMsgs + ": %d\r\n" + JSPullRequestPendingBytes + ": %d\r\n\r\n"
+	var didSend bool
 
-	// Process the consumers.
-	for _, ca := range consumers {
-		// Locate the consumer itself.
-		if acc, err := s.LookupAccount(ca.Client.Account); err == nil && acc != nil {
-			if mset, err := acc.lookupStream(ca.Stream); err == nil && mset != nil {
-				if o := mset.lookupConsumer(ca.Name); o != nil {
-					if pr := o.pendingRequestReplies(); len(pr) > 0 {
-						o.mu.Lock()
-						for _, reply := range pr {
-							o.outq.send(newJSPubMsg(reply, _EMPTY_, _EMPTY_, hdr, nil, nil, 0))
-						}
-						o.mu.Unlock()
+	for _, jsa := range js.accounts {
+		jsa.mu.RLock()
+		for _, stream := range jsa.streams {
+			stream.mu.RLock()
+			for _, o := range stream.consumers {
+				o.mu.RLock()
+				// Only signal on R1.
+				if o.cfg.Replicas <= 1 {
+					for reply, wr := range o.pendingRequests() {
+						shdr := fmt.Sprintf(hdr, wr.n, wr.b)
+						o.outq.send(newJSPubMsg(reply, _EMPTY_, _EMPTY_, []byte(shdr), nil, nil, 0))
+						didSend = true
 					}
-					state, _ := o.store.State()
-					o.deleteWithoutAdvisory()
-					js.mu.Lock()
-					// Delete old one.
-					cc.meta.ForwardProposal(encodeDeleteConsumerAssignment(ca))
-					// Encode state and new name.
-					ca.State = state
-					ca.Name = createConsumerName()
-					addEntry := encodeAddConsumerAssignmentCompressed(ca)
-					cc.meta.ForwardProposal(addEntry)
-					js.mu.Unlock()
 				}
+				o.mu.RUnlock()
 			}
+			stream.mu.RUnlock()
 		}
+		jsa.mu.RUnlock()
 	}
-
 	// Give time for migration information to make it out of our server.
-	if len(consumers) > 0 {
+	if didSend {
 		time.Sleep(50 * time.Millisecond)
 	}
 }
@@ -861,14 +874,20 @@ func (s *Server) shutdownJetStream() {
 	var _a [512]*Account
 	accounts := _a[:0]
 
-	js.mu.RLock()
+	js.mu.Lock()
 	// Collect accounts.
 	for _, jsa := range js.accounts {
 		if a := jsa.acc(); a != nil {
 			accounts = append(accounts, a)
 		}
 	}
-	js.mu.RUnlock()
+	accPurgeSub := js.accountPurge
+	js.accountPurge = nil
+	js.mu.Unlock()
+
+	if accPurgeSub != nil {
+		s.sysUnsubscribe(accPurgeSub)
+	}
 
 	for _, a := range accounts {
 		a.removeJetStream()
@@ -881,7 +900,13 @@ func (s *Server) shutdownJetStream() {
 	js.mu.Lock()
 	js.accounts = nil
 
+	var qch chan struct{}
+
 	if cc := js.cluster; cc != nil {
+		if cc.qch != nil {
+			qch = cc.qch
+			cc.qch = nil
+		}
 		js.stopUpdatesSub()
 		if cc.c != nil {
 			cc.c.closeConnection(ClientClosed)
@@ -890,6 +915,21 @@ func (s *Server) shutdownJetStream() {
 		cc.meta = nil
 	}
 	js.mu.Unlock()
+
+	// If we were clustered signal the monitor cluster go routine.
+	// We will wait for a bit for it to close.
+	// Do this without the lock.
+	if qch != nil {
+		select {
+		case qch <- struct{}{}:
+			select {
+			case <-qch:
+			case <-time.After(2 * time.Second):
+				s.Warnf("Did not receive signal for successful shutdown of cluster routine")
+			}
+		default:
+		}
+	}
 }
 
 // JetStreamConfig will return the current config. Useful if the system
@@ -937,9 +977,9 @@ func (s *Server) JetStreamReservedResources() (int64, int64, error) {
 }
 
 func (s *Server) getJetStream() *jetStream {
-	s.mu.Lock()
+	s.mu.RLock()
 	js := s.js
-	s.mu.Unlock()
+	s.mu.RUnlock()
 	return js
 }
 
@@ -993,21 +1033,21 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 		return err
 	}
 
+	sysNode := s.Node()
+
 	jsa := &jsAccount{js: js, account: a, limits: limits, streams: make(map[string]*stream), sendq: sendq, usage: make(map[string]*jsaStorage)}
-	jsa.utimer = time.AfterFunc(usageTick, jsa.sendClusterUsageUpdateTimer)
 	jsa.storeDir = filepath.Join(js.config.StoreDir, a.Name)
 
+	jsa.usageMu.Lock()
+	jsa.utimer = time.AfterFunc(usageTick, jsa.sendClusterUsageUpdateTimer)
+	// Cluster mode updates to resource usage, but we always will turn on. System internal prevents echos.
+	jsa.updatesPub = fmt.Sprintf(jsaUpdatesPubT, a.Name, sysNode)
+	jsa.updatesSub, _ = s.sysSubscribe(fmt.Sprintf(jsaUpdatesSubT, a.Name), jsa.remoteUpdateUsage)
+	jsa.usageMu.Unlock()
+
 	js.accounts[a.Name] = jsa
 	js.mu.Unlock()
 
-	sysNode := s.Node()
-
-	// Cluster mode updates to resource usage, but we always will turn on. System internal prevents echos.
-	jsa.mu.Lock()
-	jsa.updatesPub = fmt.Sprintf(jsaUpdatesPubT, a.Name, sysNode)
-	jsa.updatesSub, _ = s.sysSubscribe(fmt.Sprintf(jsaUpdatesSubT, a.Name), jsa.remoteUpdateUsage)
-	jsa.mu.Unlock()
-
 	// Stamp inside account as well.
 	a.mu.Lock()
 	a.js = jsa
@@ -1041,7 +1081,8 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 		// Just need to make sure we can write to the directory.
 		// Remove the directory will create later if needed.
 		os.RemoveAll(sdir)
-
+		// when empty remove parent directory, which may have been created as well
+		os.Remove(jsa.storeDir)
 	} else {
 		// Restore any state here.
 		s.Debugf("Recovering JetStream state for account %q", a.Name)
@@ -1056,11 +1097,11 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 		if err != nil {
 			return err
 		}
-		fis, _ := ioutil.ReadDir(tdir)
+		fis, _ := os.ReadDir(tdir)
 		for _, fi := range fis {
 			metafile := filepath.Join(tdir, fi.Name(), JetStreamMetaFile)
 			metasum := filepath.Join(tdir, fi.Name(), JetStreamMetaFileSum)
-			buf, err := ioutil.ReadFile(metafile)
+			buf, err := os.ReadFile(metafile)
 			if err != nil {
 				s.Warnf("  Error reading StreamTemplate metafile %q: %v", metasum, err)
 				continue
@@ -1069,7 +1110,7 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 				s.Warnf("  Missing StreamTemplate checksum for %q", metasum)
 				continue
 			}
-			sum, err := ioutil.ReadFile(metasum)
+			sum, err := os.ReadFile(metasum)
 			if err != nil {
 				s.Warnf("  Error reading StreamTemplate checksum %q: %v", metasum, err)
 				continue
@@ -1101,8 +1142,17 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 	}
 	var consumers []*ce
 
+	// Collect any interest policy streams to check for
+	// https://github.com/nats-io/nats-server/issues/3612
+	var ipstreams []*stream
+
+	// Remember if we should be encrypted and what cipher we think we should use.
+	encrypted := s.getOpts().JetStreamKey != _EMPTY_
+	plaintext := true
+	sc := s.getOpts().JetStreamCipher
+
 	// Now recover the streams.
-	fis, _ := ioutil.ReadDir(sdir)
+	fis, _ := os.ReadDir(sdir)
 	for _, fi := range fis {
 		mdir := filepath.Join(sdir, fi.Name())
 		key := sha256.Sum256([]byte(fi.Name()))
@@ -1116,7 +1166,7 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 			s.Warnf("  Missing stream metafile for %q", metafile)
 			continue
 		}
-		buf, err := ioutil.ReadFile(metafile)
+		buf, err := os.ReadFile(metafile)
 		if err != nil {
 			s.Warnf("  Error reading metafile %q: %v", metafile, err)
 			continue
@@ -1125,7 +1175,7 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 			s.Warnf("  Missing stream checksum file %q", metasum)
 			continue
 		}
-		sum, err := ioutil.ReadFile(metasum)
+		sum, err := os.ReadFile(metasum)
 		if err != nil {
 			s.Warnf("  Error reading Stream metafile checksum %q: %v", metasum, err)
 			continue
@@ -1137,18 +1187,40 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 			continue
 		}
 
+		// Track if we are converting ciphers.
+		var osc StoreCipher
+		var convertingCiphers bool
+
 		// Check if we are encrypted.
-		if key, err := ioutil.ReadFile(filepath.Join(mdir, JetStreamMetaFileKey)); err == nil {
+		keyFile := filepath.Join(mdir, JetStreamMetaFileKey)
+		if key, err := os.ReadFile(keyFile); err == nil {
 			s.Debugf("  Stream metafile is encrypted, reading encrypted keyfile")
-			if len(key) != metaKeySize {
+			if len(key) < minMetaKeySize {
 				s.Warnf("  Bad stream encryption key length of %d", len(key))
 				continue
 			}
 			// Decode the buffer before proceeding.
-			if buf, err = s.decryptMeta(key, buf, a.Name, fi.Name()); err != nil {
-				s.Warnf("  Error decrypting our stream metafile: %v", err)
-				continue
+			nbuf, err := s.decryptMeta(sc, key, buf, a.Name, fi.Name())
+			if err != nil {
+				// See if we are changing ciphers.
+				switch sc {
+				case ChaCha:
+					nbuf, err = s.decryptMeta(AES, key, buf, a.Name, fi.Name())
+					osc, convertingCiphers = AES, true
+				case AES:
+					nbuf, err = s.decryptMeta(ChaCha, key, buf, a.Name, fi.Name())
+					osc, convertingCiphers = ChaCha, true
+				}
+				if err != nil {
+					s.Warnf("  Error decrypting our stream metafile: %v", err)
+					continue
+				}
 			}
+			buf = nbuf
+			plaintext = false
+
+			// Remove the key file to have system regenerate with the new cipher.
+			os.Remove(keyFile)
 		}
 
 		var cfg FileStreamInfo
@@ -1163,6 +1235,9 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 			}
 		}
 
+		// We had a bug that set a default de dupe window on mirror, despite that being not a valid config
+		fixCfgMirrorWithDedupWindow(&cfg.StreamConfig)
+
 		// We had a bug that could allow subjects in that had prefix or suffix spaces. We check for that here
 		// and will patch them on the fly for now. We will warn about them.
 		var hadSubjErr bool
@@ -1189,6 +1264,17 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 			cfg.StreamConfig.Subjects = nil
 		}
 
+		s.Noticef("  Starting restore for stream '%s > %s'", a.Name, cfg.StreamConfig.Name)
+
+		// Log if we are converting from plaintext to encrypted.
+		if encrypted {
+			if plaintext {
+				s.Noticef("  Encrypting stream '%s > %s'", a.Name, cfg.StreamConfig.Name)
+			} else if convertingCiphers {
+				s.Noticef("  Converting from %s to %s for stream '%s > %s'", osc, sc, a.Name, cfg.StreamConfig.Name)
+			}
+		}
+
 		// Add in the stream.
 		mset, err := a.addStream(&cfg.StreamConfig)
 		if err != nil {
@@ -1202,13 +1288,19 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 		state := mset.state()
 		s.Noticef("  Restored %s messages for stream '%s > %s'", comma(int64(state.Msgs)), mset.accName(), mset.name())
 
+		// Collect to check for dangling messages.
+		// TODO(dlc) - Can be removed eventually.
+		if cfg.StreamConfig.Retention == InterestPolicy {
+			ipstreams = append(ipstreams, mset)
+		}
+
 		// Now do the consumers.
 		odir := filepath.Join(sdir, fi.Name(), consumerDir)
 		consumers = append(consumers, &ce{mset, odir})
 	}
 
 	for _, e := range consumers {
-		ofis, _ := ioutil.ReadDir(e.odir)
+		ofis, _ := os.ReadDir(e.odir)
 		if len(ofis) > 0 {
 			s.Noticef("  Recovering %d consumers for stream - '%s > %s'", len(ofis), e.mset.accName(), e.mset.name())
 		}
@@ -1219,7 +1311,7 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 				s.Warnf("    Missing consumer metafile %q", metafile)
 				continue
 			}
-			buf, err := ioutil.ReadFile(metafile)
+			buf, err := os.ReadFile(metafile)
 			if err != nil {
 				s.Warnf("    Error reading consumer metafile %q: %v", metafile, err)
 				continue
@@ -1230,13 +1322,25 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 			}
 
 			// Check if we are encrypted.
-			if key, err := ioutil.ReadFile(filepath.Join(e.odir, ofi.Name(), JetStreamMetaFileKey)); err == nil {
+			if key, err := os.ReadFile(filepath.Join(e.odir, ofi.Name(), JetStreamMetaFileKey)); err == nil {
 				s.Debugf("  Consumer metafile is encrypted, reading encrypted keyfile")
 				// Decode the buffer before proceeding.
-				if buf, err = s.decryptMeta(key, buf, a.Name, e.mset.name()+tsep+ofi.Name()); err != nil {
-					s.Warnf("  Error decrypting our consumer metafile: %v", err)
-					continue
+				ctxName := e.mset.name() + tsep + ofi.Name()
+				nbuf, err := s.decryptMeta(sc, key, buf, a.Name, ctxName)
+				if err != nil {
+					// See if we are changing ciphers.
+					switch sc {
+					case ChaCha:
+						nbuf, err = s.decryptMeta(AES, key, buf, a.Name, ctxName)
+					case AES:
+						nbuf, err = s.decryptMeta(ChaCha, key, buf, a.Name, ctxName)
+					}
+					if err != nil {
+						s.Warnf("  Error decrypting our consumer metafile: %v", err)
+						continue
+					}
 				}
+				buf = nbuf
 			}
 
 			var cfg FileConsumerInfo
@@ -1250,7 +1354,7 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 				// the consumer can reconnect. We will create it as a durable and switch it.
 				cfg.ConsumerConfig.Durable = ofi.Name()
 			}
-			obs, err := e.mset.addConsumer(&cfg.ConsumerConfig)
+			obs, err := e.mset.addConsumerWithAssignment(&cfg.ConsumerConfig, _EMPTY_, nil, true)
 			if err != nil {
 				s.Warnf("    Error adding consumer %q: %v", cfg.Name, err)
 				continue
@@ -1274,6 +1378,11 @@ func (a *Account) EnableJetStream(limits map[string]JetStreamAccountLimits) erro
 	// Make sure to cleanup any old remaining snapshots.
 	os.RemoveAll(filepath.Join(jsa.storeDir, snapsDir))
 
+	// Check interest policy streams for auto cleanup.
+	for _, mset := range ipstreams {
+		mset.checkForOrphanMsgs()
+	}
+
 	s.Debugf("JetStream state for account %q recovered", a.Name)
 
 	return nil
@@ -1288,9 +1397,9 @@ func (a *Account) maxBytesLimits(cfg *StreamConfig) (bool, int64) {
 	if jsa == nil {
 		return false, 0
 	}
-	jsa.mu.RLock()
+	jsa.usageMu.RLock()
 	selectedLimits, _, ok := jsa.selectLimits(cfg)
-	jsa.mu.RUnlock()
+	jsa.usageMu.RUnlock()
 	if !ok {
 		return false, 0
 	}
@@ -1392,9 +1501,9 @@ func (a *Account) UpdateJetStreamLimits(limits map[string]JetStreamAccountLimits
 	}
 
 	// Calculate the delta between what we have and what we want.
-	jsa.mu.Lock()
+	jsa.usageMu.RLock()
 	dl := diffCheckedLimits(jsa.limits, limits)
-	jsa.mu.Unlock()
+	jsa.usageMu.RUnlock()
 
 	js.mu.Lock()
 	// Check the limits against existing reservations.
@@ -1405,9 +1514,9 @@ func (a *Account) UpdateJetStreamLimits(limits map[string]JetStreamAccountLimits
 	js.mu.Unlock()
 
 	// Update
-	jsa.mu.Lock()
+	jsa.usageMu.Lock()
 	jsa.limits = limits
-	jsa.mu.Unlock()
+	jsa.usageMu.Unlock()
 
 	return nil
 }
@@ -1446,6 +1555,7 @@ func (a *Account) JetStreamUsage() JetStreamAccountStats {
 		js := jsa.js
 		js.mu.RLock()
 		jsa.mu.RLock()
+		jsa.usageMu.RLock()
 		stats.Memory, stats.Store = jsa.storageTotals()
 		stats.Domain = js.config.Domain
 		stats.API = JetStreamAPIStats{
@@ -1480,6 +1590,7 @@ func (a *Account) JetStreamUsage() JetStreamAccountStats {
 				}
 			}
 		}
+		jsa.usageMu.RUnlock()
 		if cc := jsa.js.cluster; cc != nil {
 			sas := cc.streams[aname]
 			if defaultTier {
@@ -1588,10 +1699,12 @@ func (a *Account) JetStreamEnabled() bool {
 func (jsa *jsAccount) remoteUpdateUsage(sub *subscription, c *client, _ *Account, subject, _ string, msg []byte) {
 	const usageSize = 32
 
-	jsa.mu.Lock()
+	// jsa.js.srv is immutable and guaranteed to no be nil, so no lock needed.
 	s := jsa.js.srv
+
+	jsa.usageMu.Lock()
 	if len(msg) < usageSize {
-		jsa.mu.Unlock()
+		jsa.usageMu.Unlock()
 		s.Warnf("Ignoring remote usage update with size too short")
 		return
 	}
@@ -1600,7 +1713,7 @@ func (jsa *jsAccount) remoteUpdateUsage(sub *subscription, c *client, _ *Account
 		rnode = subject[li+1:]
 	}
 	if rnode == _EMPTY_ {
-		jsa.mu.Unlock()
+		jsa.usageMu.Unlock()
 		s.Warnf("Received remote usage update with no remote node")
 		return
 	}
@@ -1658,21 +1771,21 @@ func (jsa *jsAccount) remoteUpdateUsage(sub *subscription, c *client, _ *Account
 	rUsage.err = apiErrors
 	jsa.apiTotal += apiTotal
 	jsa.apiErrors += apiErrors
-	jsa.mu.Unlock()
+	jsa.usageMu.Unlock()
 }
 
 // Updates accounting on in use memory and storage. This is called from locally
 // by the lower storage layers.
 func (jsa *jsAccount) updateUsage(tierName string, storeType StorageType, delta int64) {
-	var isClustered bool
-	// Ok to check jsa.js here w/o lock.
+	// jsa.js is immutable and cannot be nil, so ok w/o lock.
 	js := jsa.js
-	if js != nil {
-		isClustered = js.isClustered()
-	}
+	// updateUsage() may be invoked under the mset's lock, so we can't get
+	// the js' lock to check if clustered. So use this function that make
+	// use of an atomic to do the check without having data race reports.
+	isClustered := js.isClusteredNoLock()
 
-	jsa.mu.Lock()
-	defer jsa.mu.Unlock()
+	jsa.usageMu.Lock()
+	defer jsa.usageMu.Unlock()
 
 	s, ok := jsa.usage[tierName]
 	if !ok {
@@ -1694,11 +1807,11 @@ func (jsa *jsAccount) updateUsage(tierName string, storeType StorageType, delta
 	}
 }
 
-const usageTick = 1500 * time.Millisecond
+var usageTick = 1500 * time.Millisecond
 
 func (jsa *jsAccount) sendClusterUsageUpdateTimer() {
-	jsa.mu.Lock()
-	defer jsa.mu.Unlock()
+	jsa.usageMu.Lock()
+	defer jsa.usageMu.Unlock()
 	jsa.sendClusterUsageUpdate()
 	if jsa.utimer != nil {
 		jsa.utimer.Reset(usageTick)
@@ -1706,11 +1819,8 @@ func (jsa *jsAccount) sendClusterUsageUpdateTimer() {
 }
 
 // Send updates to our account usage for this server.
-// Lock should be held.
+// jsa.usageMu lock should be held.
 func (jsa *jsAccount) sendClusterUsageUpdate() {
-	if jsa.js == nil || jsa.js.srv == nil || jsa.sendq == nil {
-		return
-	}
 	// These values are absolute so we can limit send rates.
 	now := time.Now()
 	if now.Sub(jsa.lupdate) < 250*time.Millisecond {
@@ -1788,7 +1898,7 @@ func (jsa *jsAccount) jetStreamAndClustered() (*jetStream, bool) {
 	return js, js.isClustered()
 }
 
-// Read lock should be held.
+// jsa.usageMu read lock should be held.
 func (jsa *jsAccount) selectLimits(cfg *StreamConfig) (JetStreamAccountLimits, string, bool) {
 	if selectedLimits, ok := jsa.limits[_EMPTY_]; ok {
 		return selectedLimits, _EMPTY_, true
@@ -1814,7 +1924,7 @@ func (jsa *jsAccount) countStreams(tier string, cfg *StreamConfig) int {
 	return streams
 }
 
-// Lock should be held.
+// jsa.usageMu read lock (at least) should be held.
 func (jsa *jsAccount) storageTotals() (uint64, uint64) {
 	mem := uint64(0)
 	store := uint64(0)
@@ -1826,8 +1936,8 @@ func (jsa *jsAccount) storageTotals() (uint64, uint64) {
 }
 
 func (jsa *jsAccount) limitsExceeded(storeType StorageType, tierName string) (bool, *ApiError) {
-	jsa.mu.RLock()
-	defer jsa.mu.RUnlock()
+	jsa.usageMu.RLock()
+	defer jsa.usageMu.RUnlock()
 
 	selectedLimits, ok := jsa.limits[tierName]
 	if !ok {
@@ -1933,16 +2043,18 @@ func (jsa *jsAccount) delete() {
 	var ts []string
 
 	jsa.mu.Lock()
+	// The update timer and subs need to be protected by usageMu lock
+	jsa.usageMu.Lock()
 	if jsa.utimer != nil {
 		jsa.utimer.Stop()
 		jsa.utimer = nil
 	}
-
 	if jsa.updatesSub != nil && jsa.js.srv != nil {
 		s := jsa.js.srv
 		s.sysUnsubscribe(jsa.updatesSub)
 		jsa.updatesSub = nil
 	}
+	jsa.usageMu.Unlock()
 
 	for _, ms := range jsa.streams {
 		streams = append(streams, ms)
@@ -2029,11 +2141,14 @@ func (js *jetStream) sufficientResources(limits map[string]JetStreamAccountLimit
 	// Since we know if we are here we are single server mode, check the account reservations.
 	var storeReserved, memReserved int64
 	for _, jsa := range js.accounts {
-		jsa.mu.RLock()
+		if jsa.account.IsExpired() {
+			continue
+		}
+		jsa.usageMu.RLock()
 		maxMemory, maxStore := totalMaxBytes(jsa.limits)
+		jsa.usageMu.RUnlock()
 		memReserved += maxMemory
 		storeReserved += maxStore
-		jsa.mu.RUnlock()
 	}
 
 	if memReserved+totalMaxMemory > js.config.MaxMemory {
@@ -2575,5 +2690,19 @@ func validateJetStreamOptions(o *Options) error {
 	default:
 		return fmt.Errorf("expected 'no_extend' for string value, got '%s'", h)
 	}
+
+	if o.JetStreamMaxCatchup < 0 {
+		return fmt.Errorf("jetstream max catchup cannot be negative")
+	}
 	return nil
 }
+
+// We had a bug that set a default de dupe window on mirror, despite that being not a valid config
+func fixCfgMirrorWithDedupWindow(cfg *StreamConfig) {
+	if cfg == nil || cfg.Mirror == nil {
+		return
+	}
+	if cfg.Duplicates != 0 {
+		cfg.Duplicates = 0
+	}
+}
