@@ -72,6 +72,18 @@ func (ms *memStore) UpdateConfig(cfg *StreamConfig) error {
 		ms.ageChk.Stop()
 		ms.ageChk = nil
 	}
+	// Make sure to update MaxMsgsPer
+	maxp := ms.maxp
+	ms.maxp = cfg.MaxMsgsPer
+	// If the value is smaller we need to enforce that.
+	if ms.maxp != 0 && ms.maxp < maxp {
+		lm := uint64(ms.maxp)
+		for _, ss := range ms.fss {
+			if ss.Msgs > lm {
+				ms.enforcePerSubjectLimit(ss)
+			}
+		}
+	}
 	ms.mu.Unlock()
 
 	if cfg.MaxAge != 0 {
@@ -98,6 +110,9 @@ func (ms *memStore) storeRawMsg(subj string, hdr, msg []byte, seq uint64, ts int
 
 	// Check if we are discarding new messages when we reach the limit.
 	if ms.cfg.Discard == DiscardNew {
+		if asl && ms.cfg.DiscardNewPer {
+			return ErrMaxMsgsPerSubject
+		}
 		if ms.cfg.MaxMsgs > 0 && ms.state.Msgs >= uint64(ms.cfg.MaxMsgs) {
 			// If we are tracking max messages per subject and are at the limit we will replace, so this is ok.
 			if !asl {
@@ -428,6 +443,10 @@ func (ms *memStore) expireMsgs() {
 // PurgeEx will remove messages based on subject filters, sequence and number of messages to keep.
 // Will return the number of purged messages.
 func (ms *memStore) PurgeEx(subject string, sequence, keep uint64) (purged uint64, err error) {
+	if sequence > 1 && keep > 0 {
+		return 0, ErrPurgeArgMismatch
+	}
+
 	if subject == _EMPTY_ || subject == fwcs {
 		if keep == 0 && (sequence == 0 || sequence == 1) {
 			return ms.Purge()
@@ -455,7 +474,7 @@ func (ms *memStore) PurgeEx(subject string, sequence, keep uint64) (purged uint6
 			ss.Msgs -= keep
 		}
 		last := ss.Last
-		if sequence > 0 {
+		if sequence > 1 {
 			last = sequence - 1
 		}
 		ms.mu.Lock()
@@ -654,30 +673,41 @@ func (ms *memStore) LoadNextMsg(filter string, wc bool, start uint64, smp *Store
 	}
 
 	isAll := filter == _EMPTY_ || filter == fwcs
-	subs := []string{filter}
-	if wc || isAll {
-		subs = subs[:0]
-		for fsubj := range ms.fss {
-			if isAll || subjectIsSubsetMatch(fsubj, filter) {
-				subs = append(subs, fsubj)
+
+	// Skip scan of mb.fss is number of messages in the block are less than
+	// 1/2 the number of subjects in mb.fss. Or we have a wc and lots of fss entries.
+	const linearScanMaxFSS = 256
+	doLinearScan := isAll || 2*int(ms.state.LastSeq-start) < len(ms.fss) || (wc && len(ms.fss) > linearScanMaxFSS)
+
+	// Initial setup.
+	fseq, lseq := start, ms.state.LastSeq
+
+	if !doLinearScan {
+		subs := []string{filter}
+		if wc || isAll {
+			subs = subs[:0]
+			for fsubj := range ms.fss {
+				if isAll || subjectIsSubsetMatch(fsubj, filter) {
+					subs = append(subs, fsubj)
+				}
 			}
 		}
-	}
-	fseq, lseq := ms.state.LastSeq, uint64(0)
-	for _, subj := range subs {
-		ss := ms.fss[subj]
-		if ss == nil {
-			continue
+		fseq, lseq = ms.state.LastSeq, uint64(0)
+		for _, subj := range subs {
+			ss := ms.fss[subj]
+			if ss == nil {
+				continue
+			}
+			if ss.First < fseq {
+				fseq = ss.First
+			}
+			if ss.Last > lseq {
+				lseq = ss.Last
+			}
 		}
-		if ss.First < fseq {
-			fseq = ss.First
+		if fseq < start {
+			fseq = start
 		}
-		if ss.Last > lseq {
-			lseq = ss.Last
-		}
-	}
-	if fseq < start {
-		fseq = start
 	}
 
 	eq := subjectsEqual
@@ -753,7 +783,12 @@ func (ms *memStore) removeSeqPerSubject(subj string, seq uint64) {
 	if seq != ss.First {
 		return
 	}
-	// TODO(dlc) - Might want to optimize this.
+	// If we know we only have 1 msg left don't need to search for next first.
+	if ss.Msgs == 1 {
+		ss.First = ss.Last
+		return
+	}
+	// TODO(dlc) - Might want to optimize this longer term.
 	for tseq := seq + 1; tseq <= ss.Last; tseq++ {
 		if sm := ms.msgs[tseq]; sm != nil && sm.subj == subj {
 			ss.First = tseq
@@ -886,23 +921,9 @@ func (ms *memStore) isClosed() bool {
 	return ms.msgs == nil
 }
 
-func (ms *memStore) incConsumers() {
-	ms.mu.Lock()
-	ms.consumers++
-	ms.mu.Unlock()
-}
-
-func (ms *memStore) decConsumers() {
-	ms.mu.Lock()
-	if ms.consumers > 0 {
-		ms.consumers--
-	}
-	ms.mu.Unlock()
-}
-
 type consumerMemStore struct {
 	mu     sync.Mutex
-	ms     *memStore
+	ms     StreamStore
 	cfg    ConsumerConfig
 	state  ConsumerState
 	closed bool
@@ -918,8 +939,25 @@ func (ms *memStore) ConsumerStore(name string, cfg *ConsumerConfig) (ConsumerSto
 	if cfg == nil || name == _EMPTY_ {
 		return nil, fmt.Errorf("bad consumer config")
 	}
-	ms.incConsumers()
-	return &consumerMemStore{ms: ms, cfg: *cfg}, nil
+	o := &consumerMemStore{ms: ms, cfg: *cfg}
+	ms.AddConsumer(o)
+	return o, nil
+}
+
+func (ms *memStore) AddConsumer(o ConsumerStore) error {
+	ms.mu.Lock()
+	ms.consumers++
+	ms.mu.Unlock()
+	return nil
+}
+
+func (ms *memStore) RemoveConsumer(o ConsumerStore) error {
+	ms.mu.Lock()
+	if ms.consumers > 0 {
+		ms.consumers--
+	}
+	ms.mu.Unlock()
+	return nil
 }
 
 func (ms *memStore) Snapshot(_ time.Duration, _, _ bool) (*SnapshotResult, error) {
@@ -974,6 +1012,19 @@ func (o *consumerMemStore) Update(state *ConsumerState) error {
 	return nil
 }
 
+// SetStarting sets our starting stream sequence.
+func (o *consumerMemStore) SetStarting(sseq uint64) error {
+	o.mu.Lock()
+	o.state.Delivered.Stream = sseq
+	o.mu.Unlock()
+	return nil
+}
+
+// HasState returns if this store has a recorded state.
+func (o *consumerMemStore) HasState() bool {
+	return false
+}
+
 func (o *consumerMemStore) UpdateDelivered(dseq, sseq, dc uint64, ts int64) error {
 	o.mu.Lock()
 	defer o.mu.Unlock()
@@ -1107,7 +1158,7 @@ func (o *consumerMemStore) Stop() error {
 	o.closed = true
 	ms := o.ms
 	o.mu.Unlock()
-	ms.decConsumers()
+	ms.RemoveConsumer(o)
 	return nil
 }
 
@@ -1120,6 +1171,16 @@ func (o *consumerMemStore) StreamDelete() error {
 }
 
 func (o *consumerMemStore) State() (*ConsumerState, error) {
+	return o.stateWithCopy(true)
+}
+
+// This will not copy pending or redelivered, so should only be done under the
+// consumer owner's lock.
+func (o *consumerMemStore) BorrowState() (*ConsumerState, error) {
+	return o.stateWithCopy(false)
+}
+
+func (o *consumerMemStore) stateWithCopy(doCopy bool) (*ConsumerState, error) {
 	o.mu.Lock()
 	defer o.mu.Unlock()
 
@@ -1132,10 +1193,18 @@ func (o *consumerMemStore) State() (*ConsumerState, error) {
 	state.Delivered = o.state.Delivered
 	state.AckFloor = o.state.AckFloor
 	if len(o.state.Pending) > 0 {
-		state.Pending = o.copyPending()
+		if doCopy {
+			state.Pending = o.copyPending()
+		} else {
+			state.Pending = o.state.Pending
+		}
 	}
 	if len(o.state.Redelivered) > 0 {
-		state.Redelivered = o.copyRedelivered()
+		if doCopy {
+			state.Redelivered = o.copyRedelivered()
+		} else {
+			state.Redelivered = o.state.Redelivered
+		}
 	}
 	return state, nil
 }
