@@ -20,7 +20,6 @@ import (
 	"errors"
 	"fmt"
 	"hash"
-	"io/ioutil"
 	"math"
 	"math/rand"
 	"net"
@@ -137,7 +136,7 @@ type raft struct {
 	csz      int
 	qn       int
 	peers    map[string]*lps
-	removed  map[string]string
+	removed  map[string]struct{}
 	acks     map[uint64]map[string]struct{}
 	pae      map[uint64]*appendEntry
 	elect    *time.Timer
@@ -228,41 +227,25 @@ type lps struct {
 }
 
 const (
-	minElectionTimeoutDefault      = 2 * time.Second
-	maxElectionTimeoutDefault      = 5 * time.Second
+	minElectionTimeoutDefault      = 4 * time.Second
+	maxElectionTimeoutDefault      = 9 * time.Second
 	minCampaignTimeoutDefault      = 100 * time.Millisecond
 	maxCampaignTimeoutDefault      = 8 * minCampaignTimeoutDefault
-	hbIntervalDefault              = 500 * time.Millisecond
-	lostQuorumIntervalDefault      = hbIntervalDefault * 20 // 10 seconds
-	lostQuorumCheckIntervalDefault = hbIntervalDefault * 20 // 10 seconds
+	hbIntervalDefault              = 1 * time.Second
+	lostQuorumIntervalDefault      = hbIntervalDefault * 10 // 10 seconds
+	lostQuorumCheckIntervalDefault = hbIntervalDefault * 10 // 10 seconds
 
 )
 
 var (
-	minElectionTimeout time.Duration
-	maxElectionTimeout time.Duration
-	minCampaignTimeout time.Duration
-	maxCampaignTimeout time.Duration
-	hbInterval         time.Duration
-	lostQuorumInterval time.Duration
-	lostQuorumCheck    time.Duration
-)
-
-func init() {
-	setDefaultRaftTimeouts()
-}
-
-// This is called on init, but also by tests that want to tweak the settings
-// and then call this as a defer to restore the defaults.
-func setDefaultRaftTimeouts() {
 	minElectionTimeout = minElectionTimeoutDefault
 	maxElectionTimeout = maxElectionTimeoutDefault
 	minCampaignTimeout = minCampaignTimeoutDefault
 	maxCampaignTimeout = maxCampaignTimeoutDefault
-	hbInterval = hbIntervalDefault
+	hbInterval         = hbIntervalDefault
 	lostQuorumInterval = lostQuorumIntervalDefault
-	lostQuorumCheck = lostQuorumCheckIntervalDefault
-}
+	lostQuorumCheck    = lostQuorumCheckIntervalDefault
+)
 
 type RaftConfig struct {
 	Name     string
@@ -346,7 +329,7 @@ func (s *Server) bootstrapRaftNode(cfg *RaftConfig, knownPeers []string, allPeer
 	} else if stat == nil || !stat.IsDir() {
 		return fmt.Errorf("raft: storage directory is not a directory")
 	}
-	tmpfile, err := ioutil.TempFile(cfg.Store, "_test_")
+	tmpfile, err := os.CreateTemp(cfg.Store, "_test_")
 	if err != nil {
 		return fmt.Errorf("raft: storage directory is not writable")
 	}
@@ -586,6 +569,27 @@ func (s *Server) reloadDebugRaftNodes() {
 	s.rnMu.RUnlock()
 }
 
+func (s *Server) stepdownRaftNodes() {
+	if s == nil {
+		return
+	}
+	var nodes []RaftNode
+	s.rnMu.RLock()
+	if len(s.raftNodes) > 0 {
+		s.Debugf("Stepping down all leader raft nodes")
+	}
+	for _, n := range s.raftNodes {
+		if n.Leader() {
+			nodes = append(nodes, n)
+		}
+	}
+	s.rnMu.RUnlock()
+
+	for _, node := range nodes {
+		node.StepDown()
+	}
+}
+
 func (s *Server) shutdownRaftNodes() {
 	if s == nil {
 		return
@@ -601,9 +605,6 @@ func (s *Server) shutdownRaftNodes() {
 	s.rnMu.RUnlock()
 
 	for _, node := range nodes {
-		if node.Leader() {
-			node.StepDown()
-		}
 		node.Stop()
 	}
 }
@@ -710,6 +711,10 @@ func (n *raft) ProposeAddPeer(peer string) error {
 // As a leader if we are proposing to remove a peer assume its already gone.
 func (n *raft) doRemovePeerAsLeader(peer string) {
 	n.Lock()
+	if n.removed == nil {
+		n.removed = map[string]struct{}{}
+	}
+	n.removed[peer] = struct{}{}
 	if _, ok := n.peers[peer]; ok {
 		delete(n.peers, peer)
 		// We should decrease our cluster size since we are tracking this peer and the peer is most likely already gone.
@@ -975,24 +980,25 @@ func (n *raft) InstallSnapshot(data []byte) error {
 	snapDir := filepath.Join(n.sd, snapshotsDir)
 	sn := fmt.Sprintf(snapFileT, snap.lastTerm, snap.lastIndex)
 	sfile := filepath.Join(snapDir, sn)
-	// Remember our latest snapshot file.
-	n.snapfile = sfile
 
-	if err := ioutil.WriteFile(sfile, n.encodeSnapshot(snap), 0640); err != nil {
+	if err := os.WriteFile(sfile, n.encodeSnapshot(snap), 0640); err != nil {
 		n.Unlock()
-		n.setWriteErr(err)
+		// We could set write err here, but if this is a temporary situation, too many open files etc.
+		// we want to retry and snapshots are not fatal.
 		return err
 	}
 
+	// Remember our latest snapshot file.
+	n.snapfile = sfile
+
 	if _, err := n.wal.Compact(snap.lastIndex); err != nil {
 		n.Unlock()
 		n.setWriteErr(err)
 		return err
 	}
-
 	n.Unlock()
 
-	psnaps, _ := ioutil.ReadDir(snapDir)
+	psnaps, _ := os.ReadDir(snapDir)
 	// Remove any old snapshots.
 	for _, fi := range psnaps {
 		pn := fi.Name()
@@ -1028,7 +1034,7 @@ func termAndIndexFromSnapFile(sn string) (term, index uint64, err error) {
 
 func (n *raft) setupLastSnapshot() {
 	snapDir := filepath.Join(n.sd, snapshotsDir)
-	psnaps, err := ioutil.ReadDir(snapDir)
+	psnaps, err := os.ReadDir(snapDir)
 	if err != nil {
 		return
 	}
@@ -1075,8 +1081,10 @@ func (n *raft) setupLastSnapshot() {
 	n.snapfile = latest
 	snap, err := n.loadLastSnapshot()
 	if err != nil {
-		os.Remove(n.snapfile)
-		n.snapfile = _EMPTY_
+		if n.snapfile != _EMPTY_ {
+			os.Remove(n.snapfile)
+			n.snapfile = _EMPTY_
+		}
 	} else {
 		n.pindex = snap.lastIndex
 		n.pterm = snap.lastTerm
@@ -1095,7 +1103,7 @@ func (n *raft) loadLastSnapshot() (*snapshot, error) {
 	if n.snapfile == _EMPTY_ {
 		return nil, errNoSnapAvailable
 	}
-	buf, err := ioutil.ReadFile(n.snapfile)
+	buf, err := os.ReadFile(n.snapfile)
 	if err != nil {
 		n.warn("Error reading snapshot: %v", err)
 		os.Remove(n.snapfile)
@@ -1152,8 +1160,8 @@ func (n *raft) isCatchingUp() bool {
 
 // Lock should be held.
 func (n *raft) isCurrent() bool {
-	// First check if we match commit and applied.
-	if n.commit != n.applied {
+	// First check if have had activity and we match commit and applied.
+	if n.commit == 0 || n.commit != n.applied {
 		n.debug("Not current, commit %d != applied %d", n.commit, n.applied)
 		return false
 	}
@@ -1479,7 +1487,27 @@ func (n *raft) Wipe() {
 	n.Delete()
 }
 
+const (
+	raftAllSubj        = "$NRG.>"
+	raftVoteSubj       = "$NRG.V.%s"
+	raftAppendSubj     = "$NRG.AE.%s"
+	raftPropSubj       = "$NRG.P.%s"
+	raftRemovePeerSubj = "$NRG.RP.%s"
+	raftReply          = "$NRG.R.%s"
+	raftCatchupReply   = "$NRG.CR.%s"
+)
+
 // Lock should be held (due to use of random generator)
+func (n *raft) newCatchupInbox() string {
+	var b [replySuffixLen]byte
+	rn := n.prand.Int63()
+	for i, l := 0, rn; i < len(b); i++ {
+		b[i] = digits[l%base]
+		l /= base
+	}
+	return fmt.Sprintf(raftCatchupReply, b[:])
+}
+
 func (n *raft) newInbox() string {
 	var b [replySuffixLen]byte
 	rn := n.prand.Int63()
@@ -1487,18 +1515,9 @@ func (n *raft) newInbox() string {
 		b[i] = digits[l%base]
 		l /= base
 	}
-	return fmt.Sprintf(raftReplySubj, b[:])
+	return fmt.Sprintf(raftReply, b[:])
 }
 
-const (
-	raftAllSubj        = "$NRG.>"
-	raftVoteSubj       = "$NRG.V.%s"
-	raftAppendSubj     = "$NRG.AE.%s"
-	raftPropSubj       = "$NRG.P.%s"
-	raftRemovePeerSubj = "$NRG.RP.%s"
-	raftReplySubj      = "$NRG.R.%s"
-)
-
 // Our internal subscribe.
 // Lock should be held.
 func (n *raft) subscribe(subject string, cb msgHandler) (*subscription, error) {
@@ -1630,7 +1649,7 @@ func (n *raft) debug(format string, args ...interface{}) {
 
 func (n *raft) warn(format string, args ...interface{}) {
 	nf := fmt.Sprintf("RAFT [%s - %s] %s", n.id, n.group, format)
-	n.s.Warnf(nf, args...)
+	n.s.RateLimitWarnf(nf, args...)
 }
 
 func (n *raft) error(format string, args ...interface{}) {
@@ -1841,6 +1860,8 @@ func (n *raft) decodeAppendEntry(msg []byte, sub *subscription, reply string) (*
 		commit: le.Uint64(msg[16:]),
 		pterm:  le.Uint64(msg[24:]),
 		pindex: le.Uint64(msg[32:]),
+		sub:    sub,
+		reply:  reply,
 	}
 	// Decode Entries.
 	ne, ri := int(le.Uint16(msg[40:])), 42
@@ -1857,8 +1878,6 @@ func (n *raft) decodeAppendEntry(msg []byte, sub *subscription, reply string) (*
 		ae.entries = append(ae.entries, &Entry{etype, msg[ri+1 : ri+le]})
 		ri += le
 	}
-	ae.reply = reply
-	ae.sub = sub
 	ae.buf = msg
 	return ae, nil
 }
@@ -2233,9 +2252,10 @@ func (n *raft) sendSnapshotToFollower(subject string) (uint64, error) {
 	var state StreamState
 	n.wal.FastState(&state)
 	if snap.lastIndex < state.FirstSeq && state.FirstSeq != 0 {
-		snap.lastIndex = state.FirstSeq
+		snap.lastIndex = state.FirstSeq - 1
 		ae.pindex = snap.lastIndex
 	}
+
 	encoding, err := ae.encode(nil)
 	if err != nil {
 		return 0, err
@@ -2254,44 +2274,52 @@ func (n *raft) catchupFollower(ar *appendEntryResponse) {
 		delete(n.progress, ar.peer)
 		q.push(n.pindex)
 	}
+
 	// Check to make sure we have this entry.
 	start := ar.index + 1
 	var state StreamState
 	n.wal.FastState(&state)
 
-	if start < state.FirstSeq || state.Msgs == 0 && start <= state.LastSeq {
+	var didSnap bool
+	if start < state.FirstSeq || (state.Msgs == 0 && start <= state.LastSeq) {
 		n.debug("Need to send snapshot to follower")
 		if lastIndex, err := n.sendSnapshotToFollower(ar.reply); err != nil {
 			n.error("Error sending snapshot to follower [%s]: %v", ar.peer, err)
 			n.Unlock()
 			return
 		} else {
-			// If no other entries can just return here.
+			// If no other entries, we can just return here.
 			if state.Msgs == 0 {
 				n.debug("Finished catching up")
 				n.Unlock()
 				return
 			}
 			n.debug("Snapshot sent, reset first entry to %d", lastIndex)
-			start = lastIndex
+			start, didSnap = lastIndex, true
 		}
 	}
 
 	ae, err := n.loadEntry(start)
 	if err != nil {
+		n.warn("Request from follower for index [%d] possibly beyond our last index [%d] - %v", start, state.LastSeq, err)
 		ae, err = n.loadFirstEntry()
 	}
 	if err != nil || ae == nil {
-		n.debug("Could not find a starting entry: %v", err)
+		n.warn("Could not find a starting entry for catchup request: %v", err)
 		n.Unlock()
 		return
 	}
 	if ae.pindex != ar.index || ae.pterm != ar.term {
-		n.debug("Our first entry does not match request from follower")
+		n.debug("Our first entry [%d:%d] does not match request from follower [%d:%d]", ae.pterm, ae.pindex, ar.term, ar.index)
 	}
 	// Create a queue for delivering updates from responses.
 	indexUpdates := n.s.newIPQueue(fmt.Sprintf("[ACC:%s] RAFT '%s' indexUpdates", n.accName, n.group)) // of uint64
-	indexUpdates.push(ae.pindex)
+	// If we did send a snapshot above, that means we skipped ahead so bump entry pindex by 1 to start for catchup stream.
+	if didSnap {
+		indexUpdates.push(ae.pindex + 1)
+	} else {
+		indexUpdates.push(ae.pindex)
+	}
 	n.progress[ar.peer] = indexUpdates
 	n.Unlock()
 
@@ -2395,9 +2423,9 @@ func (n *raft) applyCommit(index uint64) error {
 
 			// Make sure we have our removed map.
 			if n.removed == nil {
-				n.removed = make(map[string]string)
+				n.removed = make(map[string]struct{})
 			}
-			n.removed[peer] = peer
+			n.removed[peer] = struct{}{}
 
 			if _, ok := n.peers[peer]; ok {
 				delete(n.peers, peer)
@@ -2605,17 +2633,12 @@ func (n *raft) handleAppendEntry(sub *subscription, c *client, _ *Account, subje
 	} else {
 		n.warn("AppendEntry failed to be placed on internal channel: corrupt entry")
 	}
-	n.Lock()
-	// Don't reset here if we have been asked to assume leader position.
-	if !n.lxfer {
-		n.resetElectionTimeout()
-	}
-	n.Unlock()
 }
 
 // Lock should be held.
 func (n *raft) cancelCatchup() {
 	n.debug("Canceling catchup subscription since we are now up to date")
+
 	if n.catchup != nil && n.catchup.sub != nil {
 		n.unsubscribe(n.catchup.sub)
 	}
@@ -2651,7 +2674,7 @@ func (n *raft) createCatchup(ae *appendEntry) string {
 		pindex: n.pindex,
 		active: time.Now(),
 	}
-	inbox := n.newInbox()
+	inbox := n.newCatchupInbox()
 	sub, _ := n.subscribe(inbox, n.handleAppendEntry)
 	n.catchup.sub = sub
 
@@ -2659,11 +2682,23 @@ func (n *raft) createCatchup(ae *appendEntry) string {
 }
 
 // Truncate our WAL and reset.
-func (n *raft) truncateWAL(pterm, pindex uint64) {
+// Lock should be held.
+func (n *raft) truncateWAL(term, index uint64) {
 	n.debug("Truncating and repairing WAL")
-	n.pterm, n.pindex = pterm, pindex
-	if err := n.wal.Truncate(pindex); err != nil {
+
+	if err := n.wal.Truncate(index); err != nil {
 		n.setWriteErrLocked(err)
+		return
+	}
+
+	// Set after we know we have truncated properly.
+	n.pterm, n.pindex = term, index
+
+	// Check to see if we invalidated any snapshots that might have held state
+	// from the entries we are truncating.
+	if snap, _ := n.loadLastSnapshot(); snap != nil && snap.lastIndex > index {
+		os.Remove(n.snapfile)
+		n.snapfile = _EMPTY_
 	}
 }
 
@@ -2679,6 +2714,11 @@ func (n *raft) updateLeader(newLeader string) {
 func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 	n.Lock()
 
+	// Don't reset here if we have been asked to assume leader position.
+	if !n.lxfer {
+		n.resetElectionTimeout()
+	}
+
 	// Just return if closed or we had previous write error.
 	if n.state == Closed || n.werr != nil {
 		n.Unlock()
@@ -2700,11 +2740,12 @@ func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 		} else {
 			// Let them know we are the leader.
 			ar := &appendEntryResponse{n.term, n.pindex, n.id, false, _EMPTY_}
-			n.Unlock()
 			n.debug("AppendEntry ignoring old term from another leader")
 			n.sendRPC(ae.reply, _EMPTY_, ar.encode(arbuf))
-			return
 		}
+		// Always return here from processing.
+		n.Unlock()
+		return
 	}
 
 	// If we received an append entry as a candidate we should convert to a follower.
@@ -2754,6 +2795,8 @@ func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 		if cs := n.catchup; cs != nil && n.pterm >= cs.cterm && n.pindex >= cs.cindex {
 			// If we are here we are good, so if we have a catchup pending we can cancel.
 			n.cancelCatchup()
+			// Reset our notion of catching up.
+			catchingUp = false
 		} else if isNew {
 			var ar *appendEntryResponse
 			var inbox string
@@ -2796,20 +2839,30 @@ func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 	if ae.pterm != n.pterm || ae.pindex != n.pindex {
 		// Check if this is a lower index than what we were expecting.
 		if ae.pindex < n.pindex {
+			n.debug("AppendEntry detected pindex less than ours: %d:%d vs %d:%d", ae.pterm, ae.pindex, n.pterm, n.pindex)
 			var ar *appendEntryResponse
-			if eae, err := n.loadEntry(ae.pindex); err == nil && eae != nil {
-				// If terms mismatched, delete that entry and all others past it.
-				if ae.pterm > eae.pterm {
+
+			var success bool
+			eae, err := n.loadEntry(ae.pindex)
+			// If terms mismatched, or we got an error loading, delete that entry and all others past it.
+			if eae != nil && ae.pterm > eae.pterm || err != nil {
+				// Truncate will reset our pterm and pindex. Only do so if we have an entry.
+				if eae != nil {
 					n.truncateWAL(ae.pterm, ae.pindex)
-					ar = &appendEntryResponse{n.pterm, n.pindex, n.id, false, _EMPTY_}
-				} else {
-					ar = &appendEntryResponse{ae.pterm, ae.pindex, n.id, true, _EMPTY_}
 				}
+				// Make sure to cancel any catchups in progress.
+				if catchingUp {
+					n.cancelCatchup()
+				}
+			} else {
+				// Inherit regardless.
+				n.pterm = ae.pterm
+				success = true
 			}
+			// Create response.
+			ar = &appendEntryResponse{ae.pterm, ae.pindex, n.id, success, _EMPTY_}
 			n.Unlock()
-			if ar != nil {
-				n.sendRPC(ae.reply, _EMPTY_, ar.encode(arbuf))
-			}
+			n.sendRPC(ae.reply, _EMPTY_, ar.encode(arbuf))
 			return
 		}
 
@@ -2826,7 +2879,8 @@ func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 				n.Unlock()
 				return
 			}
-			// Snapshots and peerstate will always be together when a leader is catching us up.
+			// This means we already entered into a catchup state but what the leader sent us did not match what we expected.
+			// Snapshots and peerstate will always be together when a leader is catching us up in this fashion.
 			if len(ae.entries) != 2 || ae.entries[0].Type != EntrySnapshot || ae.entries[1].Type != EntryPeerState {
 				n.warn("Expected first catchup entry to be a snapshot and peerstate, will retry")
 				n.cancelCatchup()
@@ -2844,6 +2898,7 @@ func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 				n.Unlock()
 				return
 			}
+
 			n.pindex = ae.pindex
 			n.pterm = ae.pterm
 			n.commit = ae.pindex
@@ -2887,7 +2942,7 @@ func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 			}
 			// Save in memory for faster processing during applyCommit.
 			n.pae[n.pindex] = ae
-			if l := len(n.pae); l > paeWarnThreshold && l%1000 == 0 {
+			if l := len(n.pae); l > paeWarnThreshold && l%paeWarnModulo == 0 {
 				n.warn("%d append entries pending", len(n.pae))
 			}
 
@@ -2935,11 +2990,16 @@ func (n *raft) processAppendEntry(ae *appendEntry, sub *subscription) {
 		}
 	}
 
-	ar := appendEntryResponse{n.pterm, n.pindex, n.id, true, _EMPTY_}
+	var ar *appendEntryResponse
+	if sub != nil {
+		ar = &appendEntryResponse{n.pterm, n.pindex, n.id, true, _EMPTY_}
+	}
 	n.Unlock()
 
 	// Success. Send our response.
-	n.sendRPC(ae.reply, _EMPTY_, ar.encode(arbuf))
+	if ar != nil {
+		n.sendRPC(ae.reply, _EMPTY_, ar.encode(arbuf))
+	}
 }
 
 // Lock should be held.
@@ -3016,17 +3076,11 @@ func (n *raft) storeToWAL(ae *appendEntry) error {
 
 	// Sanity checking for now.
 	if index := ae.pindex + 1; index != seq {
-		// We are missing store state from our state.
+		n.warn("Wrong index, ae is %+v, index stored was %d, n.pindex is %d", ae, seq, n.pindex)
 		if index > seq {
-			// Reset to last before this one.
-			if ae, err := n.loadEntry(seq - 1); err == nil && ae != nil {
-				nl := n.selectNextLeader()
-				n.truncateWAL(ae.pterm, ae.pindex)
-				if n.state == Leader {
-					n.stepdown.push(nl)
-				}
-			} else {
-				panic(fmt.Sprintf("[%s | %s] Wrong index, ae is %+v, seq is %d, n.pindex is %d\n\n", n.s, n.group, ae, seq, n.pindex))
+			// We are missing store state from our state. We need to stepdown at this point.
+			if n.state == Leader {
+				n.stepdown.push(n.selectNextLeader())
 			}
 		} else {
 			// Truncate back to our last known.
@@ -3040,7 +3094,8 @@ func (n *raft) storeToWAL(ae *appendEntry) error {
 	return nil
 }
 
-const paeWarnThreshold = 32 * 1024
+const paeWarnThreshold = 20_000
+const paeWarnModulo = 5_000
 
 func (n *raft) sendAppendEntry(entries []*Entry) {
 	n.Lock()
@@ -3065,7 +3120,7 @@ func (n *raft) sendAppendEntry(entries []*Entry) {
 
 		// Save in memory for faster processing during applyCommit.
 		n.pae[n.pindex] = ae
-		if l := len(n.pae); l > paeWarnThreshold && l%1000 == 0 {
+		if l := len(n.pae); l > paeWarnThreshold && l%paeWarnModulo == 0 {
 			n.warn("%d append entries pending", len(n.pae))
 		}
 	}
@@ -3180,15 +3235,13 @@ func (n *raft) decodeVoteRequest(msg []byte, reply string) *voteRequest {
 	if len(msg) != voteRequestLen {
 		return nil
 	}
-	// Need to copy for now b/c of candidate.
-	msg = copyBytes(msg)
 
 	var le = binary.LittleEndian
 	return &voteRequest{
 		term:      le.Uint64(msg[0:]),
 		lastTerm:  le.Uint64(msg[8:]),
 		lastIndex: le.Uint64(msg[16:]),
-		candidate: string(msg[24 : 24+idLen]),
+		candidate: string(copyBytes(msg[24 : 24+idLen])),
 		reply:     reply,
 	}
 }
@@ -3215,14 +3268,14 @@ func writePeerState(sd string, ps *peerState) error {
 	if _, err := os.Stat(psf); err != nil && !os.IsNotExist(err) {
 		return err
 	}
-	if err := ioutil.WriteFile(psf, encodePeerState(ps), 0640); err != nil {
+	if err := os.WriteFile(psf, encodePeerState(ps), 0640); err != nil {
 		return err
 	}
 	return nil
 }
 
 func readPeerState(sd string) (ps *peerState, err error) {
-	buf, err := ioutil.ReadFile(filepath.Join(sd, peerStateFile))
+	buf, err := os.ReadFile(filepath.Join(sd, peerStateFile))
 	if err != nil {
 		return nil, err
 	}
@@ -3235,7 +3288,7 @@ const termVoteLen = idLen + 8
 // readTermVote will read the largest term and who we voted from to stable storage.
 // Lock should be held.
 func (n *raft) readTermVote() (term uint64, voted string, err error) {
-	buf, err := ioutil.ReadFile(filepath.Join(n.sd, termVoteFile))
+	buf, err := os.ReadFile(filepath.Join(n.sd, termVoteFile))
 	if err != nil {
 		return 0, noVote, err
 	}
@@ -3251,7 +3304,7 @@ func (n *raft) readTermVote() (term uint64, voted string, err error) {
 // Lock should be held.
 func (n *raft) setWriteErrLocked(err error) {
 	// Ignore if already set.
-	if n.werr == err {
+	if n.werr == err || err == nil {
 		return
 	}
 	// Ignore non-write errors.
@@ -3298,7 +3351,7 @@ func (n *raft) fileWriter() {
 			copy(buf[0:], n.wtv)
 			n.RUnlock()
 			<-dios
-			err := ioutil.WriteFile(tvf, buf[:], 0640)
+			err := os.WriteFile(tvf, buf[:], 0640)
 			dios <- struct{}{}
 			if err != nil {
 				n.setWriteErr(err)
@@ -3309,7 +3362,7 @@ func (n *raft) fileWriter() {
 			buf := copyBytes(n.wps)
 			n.RUnlock()
 			<-dios
-			err := ioutil.WriteFile(psf, buf, 0640)
+			err := os.WriteFile(psf, buf, 0640)
 			dios <- struct{}{}
 			if err != nil {
 				n.setWriteErr(err)
@@ -3579,13 +3632,24 @@ func (n *raft) switchToCandidate() {
 
 func (n *raft) switchToLeader() {
 	n.Lock()
-	defer n.Unlock()
 	if n.state == Closed {
+		n.Unlock()
 		return
 	}
 	n.debug("Switching to leader")
 
+	var state StreamState
+	n.wal.FastState(&state)
+
+	// Check if we have items pending as we are taking over.
+	sendHB := state.LastSeq > n.commit
+
 	n.lxfer = false
 	n.updateLeader(n.id)
 	n.switchState(Leader)
+	n.Unlock()
+
+	if sendHB {
+		n.sendHeartbeat()
+	}
 }
