@@ -1,4 +1,4 @@
-// Copyright 2016-2019 The NATS Authors
+// Copyright 2016-2020 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -29,20 +29,22 @@ import (
 // Common byte variables for wildcards and token separator.
 const (
 	pwc   = '*'
+	pwcs  = "*"
 	fwc   = '>'
+	fwcs  = ">"
 	tsep  = "."
 	btsep = '.'
 )
 
 // Sublist related errors
 var (
-	ErrInvalidSubject = errors.New("sublist: invalid subject")
-	ErrNotFound       = errors.New("sublist: no matches found")
+	ErrInvalidSubject    = errors.New("sublist: invalid subject")
+	ErrNotFound          = errors.New("sublist: no matches found")
+	ErrNilChan           = errors.New("sublist: nil channel")
+	ErrAlreadyRegistered = errors.New("sublist: notification already registered")
 )
 
 const (
-	// slNoCache for cacheNum means cache is disabled.
-	slNoCache = -22
 	// cacheMax is used to bound limit the frontend cache
 	slCacheMax = 1024
 	// If we run a sweeper we will drain to this count.
@@ -66,12 +68,19 @@ type Sublist struct {
 	inserts   uint64
 	removes   uint64
 	root      *level
-	cache     *sync.Map
-	cacheNum  int32
+	cache     map[string]*SublistResult
 	ccSweep   int32
+	notify    *notifyMaps
 	count     uint32
 }
 
+// notifyMaps holds maps of arrays of channels for notifications
+// on a change of interest.
+type notifyMaps struct {
+	insert map[string][]chan<- bool
+	remove map[string][]chan<- bool
+}
+
 // A node contains subscriptions and a pointer to the next level.
 type node struct {
 	next  *level
@@ -105,9 +114,9 @@ func newLevel() *level {
 // NewSublist will create a default sublist with caching enabled per the flag.
 func NewSublist(enableCache bool) *Sublist {
 	if enableCache {
-		return &Sublist{root: newLevel(), cache: &sync.Map{}}
+		return &Sublist{root: newLevel(), cache: make(map[string]*SublistResult)}
 	}
-	return &Sublist{root: newLevel(), cacheNum: slNoCache}
+	return &Sublist{root: newLevel()}
 }
 
 // NewSublistWithCache will create a default sublist with caching enabled.
@@ -122,7 +131,223 @@ func NewSublistNoCache() *Sublist {
 
 // CacheEnabled returns whether or not caching is enabled for this sublist.
 func (s *Sublist) CacheEnabled() bool {
-	return atomic.LoadInt32(&s.cacheNum) != slNoCache
+	s.RLock()
+	enabled := s.cache != nil
+	s.RUnlock()
+	return enabled
+}
+
+// RegisterNotification will register for notifications when interest for the given
+// subject changes. The subject must be a literal publish type subject.
+// The notification is true for when the first interest for a subject is inserted,
+// and false when all interest in the subject is removed. Note that this interest
+// needs to be exact and that wildcards will not trigger the notifications. The sublist
+// will not block when trying to send the notification. Its up to the caller to make
+// sure the channel send will not block.
+func (s *Sublist) RegisterNotification(subject string, notify chan<- bool) error {
+	return s.registerNotification(subject, _EMPTY_, notify)
+}
+
+func (s *Sublist) RegisterQueueNotification(subject, queue string, notify chan<- bool) error {
+	return s.registerNotification(subject, queue, notify)
+}
+
+func (s *Sublist) registerNotification(subject, queue string, notify chan<- bool) error {
+	if subjectHasWildcard(subject) {
+		return ErrInvalidSubject
+	}
+	if notify == nil {
+		return ErrNilChan
+	}
+
+	var hasInterest bool
+	r := s.Match(subject)
+
+	if len(r.psubs)+len(r.qsubs) > 0 {
+		if queue == _EMPTY_ {
+			for _, sub := range r.psubs {
+				if string(sub.subject) == subject {
+					hasInterest = true
+					break
+				}
+			}
+		} else {
+			for _, qsub := range r.qsubs {
+				qs := qsub[0]
+				if string(qs.subject) == subject && string(qs.queue) == queue {
+					hasInterest = true
+					break
+				}
+			}
+		}
+	}
+
+	key := keyFromSubjectAndQueue(subject, queue)
+	var err error
+
+	s.Lock()
+	if s.notify == nil {
+		s.notify = &notifyMaps{
+			insert: make(map[string][]chan<- bool),
+			remove: make(map[string][]chan<- bool),
+		}
+	}
+	// Check which list to add us to.
+	if hasInterest {
+		err = s.addRemoveNotify(key, notify)
+	} else {
+		err = s.addInsertNotify(key, notify)
+	}
+	s.Unlock()
+
+	if err == nil {
+		sendNotification(notify, hasInterest)
+	}
+	return err
+}
+
+// Lock should be held.
+func chkAndRemove(key string, notify chan<- bool, ms map[string][]chan<- bool) bool {
+	chs := ms[key]
+	for i, ch := range chs {
+		if ch == notify {
+			chs[i] = chs[len(chs)-1]
+			chs = chs[:len(chs)-1]
+			if len(chs) == 0 {
+				delete(ms, key)
+			}
+			return true
+		}
+	}
+	return false
+}
+
+func (s *Sublist) ClearNotification(subject string, notify chan<- bool) bool {
+	return s.clearNotification(subject, _EMPTY_, notify)
+}
+
+func (s *Sublist) ClearQueueNotification(subject, queue string, notify chan<- bool) bool {
+	return s.clearNotification(subject, queue, notify)
+}
+
+func (s *Sublist) clearNotification(subject, queue string, notify chan<- bool) bool {
+	s.Lock()
+	if s.notify == nil {
+		s.Unlock()
+		return false
+	}
+	key := keyFromSubjectAndQueue(subject, queue)
+	// Check both, start with remove.
+	didRemove := chkAndRemove(key, notify, s.notify.remove)
+	didRemove = didRemove || chkAndRemove(key, notify, s.notify.insert)
+	// Check if everything is gone
+	if len(s.notify.remove)+len(s.notify.insert) == 0 {
+		s.notify = nil
+	}
+	s.Unlock()
+	return didRemove
+}
+
+func sendNotification(ch chan<- bool, hasInterest bool) {
+	select {
+	case ch <- hasInterest:
+	default:
+	}
+}
+
+// Add a new channel for notification in insert map.
+// Write lock should be held.
+func (s *Sublist) addInsertNotify(subject string, notify chan<- bool) error {
+	return s.addNotify(s.notify.insert, subject, notify)
+}
+
+// Add a new channel for notification in removal map.
+// Write lock should be held.
+func (s *Sublist) addRemoveNotify(subject string, notify chan<- bool) error {
+	return s.addNotify(s.notify.remove, subject, notify)
+}
+
+// Add a new channel for notification.
+// Write lock should be held.
+func (s *Sublist) addNotify(m map[string][]chan<- bool, subject string, notify chan<- bool) error {
+	chs := m[subject]
+	if len(chs) > 0 {
+		// Check to see if this chan is already registered.
+		for _, ch := range chs {
+			if ch == notify {
+				return ErrAlreadyRegistered
+			}
+		}
+	}
+
+	m[subject] = append(chs, notify)
+	return nil
+}
+
+// To generate a key from subject and queue. We just add spc.
+func keyFromSubjectAndQueue(subject, queue string) string {
+	if len(queue) == 0 {
+		return subject
+	}
+	var sb strings.Builder
+	sb.WriteString(subject)
+	sb.WriteString(" ")
+	sb.WriteString(queue)
+	return sb.String()
+}
+
+// chkForInsertNotification will check to see if we need to notify on this subject.
+// Write lock should be held.
+func (s *Sublist) chkForInsertNotification(subject, queue string) {
+	key := keyFromSubjectAndQueue(subject, queue)
+
+	// All notify subjects are also literal so just do a hash lookup here.
+	if chs := s.notify.insert[key]; len(chs) > 0 {
+		for _, ch := range chs {
+			sendNotification(ch, true)
+		}
+		// Move from the insert map to the remove map.
+		s.notify.remove[key] = append(s.notify.remove[key], chs...)
+		delete(s.notify.insert, key)
+	}
+}
+
+// chkForRemoveNotification will check to see if we need to notify on this subject.
+// Write lock should be held.
+func (s *Sublist) chkForRemoveNotification(subject, queue string) {
+	key := keyFromSubjectAndQueue(subject, queue)
+	if chs := s.notify.remove[key]; len(chs) > 0 {
+		// We need to always check that we have no interest anymore.
+		var hasInterest bool
+		r := s.matchNoLock(subject)
+
+		if len(r.psubs)+len(r.qsubs) > 0 {
+			if queue == _EMPTY_ {
+				for _, sub := range r.psubs {
+					if string(sub.subject) == subject {
+						hasInterest = true
+						break
+					}
+				}
+			} else {
+				for _, qsub := range r.qsubs {
+					qs := qsub[0]
+					if string(qs.subject) == subject && string(qs.queue) == queue {
+						hasInterest = true
+						break
+					}
+				}
+			}
+		}
+		if !hasInterest {
+			for _, ch := range chs {
+				sendNotification(ch, false)
+			}
+			// Move from the remove map to the insert map.
+			s.notify.insert[key] = append(s.notify.insert[key], chs...)
+			delete(s.notify.remove, key)
+		}
+	}
 }
 
 // Insert adds a subscription into the sublist
@@ -142,9 +367,9 @@ func (s *Sublist) Insert(sub *subscription) error {
 
 	s.Lock()
 
-	sfwc := false
-	l := s.root
+	var sfwc, haswc, isnew bool
 	var n *node
+	l := s.root
 
 	for _, t := range tokens {
 		lt := len(t)
@@ -159,9 +384,10 @@ func (s *Sublist) Insert(sub *subscription) error {
 			switch t[0] {
 			case pwc:
 				n = l.pwc
+				haswc = true
 			case fwc:
 				n = l.fwc
-				sfwc = true
+				haswc, sfwc = true, true
 			default:
 				n = l.nodes[t]
 			}
@@ -188,6 +414,7 @@ func (s *Sublist) Insert(sub *subscription) error {
 	}
 	if sub.queue == nil {
 		n.psubs[sub] = sub
+		isnew = len(n.psubs) == 1
 		if n.plist != nil {
 			n.plist = append(n.plist, sub)
 		} else if len(n.psubs) > plistMin {
@@ -207,6 +434,7 @@ func (s *Sublist) Insert(sub *subscription) error {
 		if !ok {
 			subs = make(map[*subscription]*subscription)
 			n.qsubs[qname] = subs
+			isnew = true
 		}
 		subs[sub] = sub
 	}
@@ -217,7 +445,11 @@ func (s *Sublist) Insert(sub *subscription) error {
 	s.addToCache(subject, sub)
 	atomic.AddUint64(&s.genid, 1)
 
+	if s.notify != nil && isnew && !haswc && len(s.notify.insert) > 0 {
+		s.chkForInsertNotification(subject, string(sub.queue))
+	}
 	s.Unlock()
+
 	return nil
 }
 
@@ -250,26 +482,23 @@ func (r *SublistResult) addSubToResult(sub *subscription) *SublistResult {
 
 // addToCache will add the new entry to the existing cache
 // entries if needed. Assumes write lock is held.
+// Assumes write lock is held.
 func (s *Sublist) addToCache(subject string, sub *subscription) {
 	if s.cache == nil {
 		return
 	}
 	// If literal we can direct match.
 	if subjectIsLiteral(subject) {
-		if v, ok := s.cache.Load(subject); ok {
-			r := v.(*SublistResult)
-			s.cache.Store(subject, r.addSubToResult(sub))
+		if r := s.cache[subject]; r != nil {
+			s.cache[subject] = r.addSubToResult(sub)
 		}
 		return
 	}
-	s.cache.Range(func(k, v interface{}) bool {
-		key := k.(string)
-		r := v.(*SublistResult)
+	for key, r := range s.cache {
 		if matchLiteral(key, subject) {
-			s.cache.Store(key, r.addSubToResult(sub))
+			s.cache[key] = r.addSubToResult(sub)
 		}
-		return true
-	})
+	}
 }
 
 // removeFromCache will remove the sub from any active cache entries.
@@ -280,23 +509,15 @@ func (s *Sublist) removeFromCache(subject string, sub *subscription) {
 	}
 	// If literal we can direct match.
 	if subjectIsLiteral(subject) {
-		// Load for accounting
-		if _, ok := s.cache.Load(subject); ok {
-			s.cache.Delete(subject)
-			atomic.AddInt32(&s.cacheNum, -1)
-		}
+		delete(s.cache, subject)
 		return
 	}
-	s.cache.Range(func(k, v interface{}) bool {
-		key := k.(string)
+	// Wildcard here.
+	for key := range s.cache {
 		if matchLiteral(key, subject) {
-			// Since someone else may be referecing, can't modify the list
-			// safely, just let it re-populate.
-			s.cache.Delete(key)
-			atomic.AddInt32(&s.cacheNum, -1)
+			delete(s.cache, key)
 		}
-		return true
-	})
+	}
 }
 
 // a place holder for an empty result.
@@ -305,14 +526,27 @@ var emptyResult = &SublistResult{}
 // Match will match all entries to the literal subject.
 // It will return a set of results for both normal and queue subscribers.
 func (s *Sublist) Match(subject string) *SublistResult {
+	return s.match(subject, true)
+}
+
+func (s *Sublist) matchNoLock(subject string) *SublistResult {
+	return s.match(subject, false)
+}
+
+func (s *Sublist) match(subject string, doLock bool) *SublistResult {
 	atomic.AddUint64(&s.matches, 1)
 
 	// Check cache first.
-	if atomic.LoadInt32(&s.cacheNum) > 0 {
-		if r, ok := s.cache.Load(subject); ok {
-			atomic.AddUint64(&s.cacheHits, 1)
-			return r.(*SublistResult)
-		}
+	if doLock {
+		s.RLock()
+	}
+	r, ok := s.cache[subject]
+	if doLock {
+		s.RUnlock()
+	}
+	if ok {
+		atomic.AddUint64(&s.cacheHits, 1)
+		return r
 	}
 
 	tsa := [32]string{}
@@ -320,10 +554,16 @@ func (s *Sublist) Match(subject string) *SublistResult {
 	start := 0
 	for i := 0; i < len(subject); i++ {
 		if subject[i] == btsep {
+			if i-start == 0 {
+				return emptyResult
+			}
 			tokens = append(tokens, subject[start:i])
 			start = i + 1
 		}
 	}
+	if start >= len(subject) {
+		return emptyResult
+	}
 	tokens = append(tokens, subject[start:])
 
 	// FIXME(dlc) - Make shared pool between sublist and client readLoop?
@@ -331,19 +571,24 @@ func (s *Sublist) Match(subject string) *SublistResult {
 
 	// Get result from the main structure and place into the shared cache.
 	// Hold the read lock to avoid race between match and store.
-	var n int32
+	var n int
+
+	if doLock {
+		s.Lock()
+	}
 
-	s.RLock()
 	matchLevel(s.root, tokens, result)
 	// Check for empty result.
 	if len(result.psubs) == 0 && len(result.qsubs) == 0 {
 		result = emptyResult
 	}
 	if s.cache != nil {
-		s.cache.Store(subject, result)
-		n = atomic.AddInt32(&s.cacheNum, 1)
+		s.cache[subject] = result
+		n = len(s.cache)
+	}
+	if doLock {
+		s.Unlock()
 	}
-	s.RUnlock()
 
 	// Reduce the cache count if we have exceeded our set maximum.
 	if n > slCacheMax && atomic.CompareAndSwapInt32(&s.ccSweep, 0, 1) {
@@ -359,11 +604,12 @@ func (s *Sublist) reduceCacheCount() {
 	defer atomic.StoreInt32(&s.ccSweep, 0)
 	// If we are over the cache limit randomly drop until under the limit.
 	s.Lock()
-	s.cache.Range(func(k, v interface{}) bool {
-		s.cache.Delete(k.(string))
-		n := atomic.AddInt32(&s.cacheNum, -1)
-		return n >= slCacheSweep
-	})
+	for key := range s.cache {
+		delete(s.cache, key)
+		if len(s.cache) <= slCacheSweep {
+			break
+		}
+	}
 	s.Unlock()
 }
 
@@ -472,7 +718,7 @@ type lnt struct {
 }
 
 // Raw low level remove, can do batches with lock held outside.
-func (s *Sublist) remove(sub *subscription, shouldLock bool) error {
+func (s *Sublist) remove(sub *subscription, shouldLock bool, doCacheUpdates bool) error {
 	subject := string(sub.subject)
 	tsa := [32]string{}
 	tokens := tsa[:0]
@@ -490,9 +736,9 @@ func (s *Sublist) remove(sub *subscription, shouldLock bool) error {
 		defer s.Unlock()
 	}
 
-	sfwc := false
-	l := s.root
+	var sfwc, haswc bool
 	var n *node
+	l := s.root
 
 	// Track levels for pruning
 	var lnts [32]lnt
@@ -512,9 +758,10 @@ func (s *Sublist) remove(sub *subscription, shouldLock bool) error {
 			switch t[0] {
 			case pwc:
 				n = l.pwc
+				haswc = true
 			case fwc:
 				n = l.fwc
-				sfwc = true
+				haswc, sfwc = true, true
 			default:
 				n = l.nodes[t]
 			}
@@ -526,7 +773,8 @@ func (s *Sublist) remove(sub *subscription, shouldLock bool) error {
 			l = nil
 		}
 	}
-	if !s.removeFromNode(n, sub) {
+	removed, last := s.removeFromNode(n, sub)
+	if !removed {
 		return ErrNotFound
 	}
 
@@ -539,81 +787,52 @@ func (s *Sublist) remove(sub *subscription, shouldLock bool) error {
 			l.pruneNode(n, t)
 		}
 	}
-	s.removeFromCache(subject, sub)
-	atomic.AddUint64(&s.genid, 1)
+	if doCacheUpdates {
+		s.removeFromCache(subject, sub)
+		atomic.AddUint64(&s.genid, 1)
+	}
+
+	if s.notify != nil && last && !haswc && len(s.notify.remove) > 0 {
+		s.chkForRemoveNotification(subject, string(sub.queue))
+	}
 
 	return nil
 }
 
 // Remove will remove a subscription.
 func (s *Sublist) Remove(sub *subscription) error {
-	return s.remove(sub, true)
+	return s.remove(sub, true, true)
 }
 
 // RemoveBatch will remove a list of subscriptions.
 func (s *Sublist) RemoveBatch(subs []*subscription) error {
+	if len(subs) == 0 {
+		return nil
+	}
+
 	s.Lock()
 	defer s.Unlock()
 
+	// TODO(dlc) - We could try to be smarter here for a client going away but the account
+	// has a large number of subscriptions compared to this client. Quick and dirty testing
+	// though said just disabling all the time best for now.
+
+	// Turn off our cache if enabled.
+	wasEnabled := s.cache != nil
+	s.cache = nil
 	for _, sub := range subs {
-		if err := s.remove(sub, false); err != nil {
+		if err := s.remove(sub, false, false); err != nil {
 			return err
 		}
 	}
+	// Turn caching back on here.
+	atomic.AddUint64(&s.genid, 1)
+	if wasEnabled {
+		s.cache = make(map[string]*SublistResult)
+	}
 	return nil
 }
 
-func (s *Sublist) checkNodeForClientSubs(n *node, c *client) {
-	var removed uint32
-	for _, sub := range n.psubs {
-		if sub.client == c {
-			if s.removeFromNode(n, sub) {
-				s.removeFromCache(string(sub.subject), sub)
-				removed++
-			}
-		}
-	}
-	// Queue subscriptions
-	for _, qr := range n.qsubs {
-		for _, sub := range qr {
-			if sub.client == c {
-				if s.removeFromNode(n, sub) {
-					s.removeFromCache(string(sub.subject), sub)
-					removed++
-				}
-			}
-		}
-	}
-	s.count -= removed
-	s.removes += uint64(removed)
-}
-
-func (s *Sublist) removeClientSubs(l *level, c *client) {
-	for _, n := range l.nodes {
-		s.checkNodeForClientSubs(n, c)
-		s.removeClientSubs(n.next, c)
-	}
-	if l.pwc != nil {
-		s.checkNodeForClientSubs(l.pwc, c)
-		s.removeClientSubs(l.pwc.next, c)
-	}
-	if l.fwc != nil {
-		s.checkNodeForClientSubs(l.fwc, c)
-		s.removeClientSubs(l.fwc.next, c)
-	}
-}
-
-// RemoveAllForClient will remove all subscriptions for a given client.
-func (s *Sublist) RemoveAllForClient(c *client) {
-	s.Lock()
-	removes := s.removes
-	s.removeClientSubs(s.root, c)
-	if s.removes != removes {
-		atomic.AddUint64(&s.genid, 1)
-	}
-	s.Unlock()
-}
-
 // pruneNode is used to prune an empty node from the tree.
 func (l *level) pruneNode(n *node, t string) {
 	if n == nil {
@@ -652,9 +871,9 @@ func (l *level) numNodes() int {
 }
 
 // Remove the sub for the given node.
-func (s *Sublist) removeFromNode(n *node, sub *subscription) (found bool) {
+func (s *Sublist) removeFromNode(n *node, sub *subscription) (found, last bool) {
 	if n == nil {
-		return false
+		return false, true
 	}
 	if sub.queue == nil {
 		_, found = n.psubs[sub]
@@ -665,7 +884,7 @@ func (s *Sublist) removeFromNode(n *node, sub *subscription) (found bool) {
 			// to Match as needed.
 			n.plist = nil
 		}
-		return found
+		return found, len(n.psubs) == 0
 	}
 
 	// We have a queue group subscription here
@@ -673,9 +892,12 @@ func (s *Sublist) removeFromNode(n *node, sub *subscription) (found bool) {
 	_, found = qsub[sub]
 	delete(qsub, sub)
 	if len(qsub) == 0 {
+		// This is the last queue subscription interest when len(qsub) == 0, not
+		// when n.qsubs is empty.
+		last = true
 		delete(n.qsubs, string(sub.queue))
 	}
-	return found
+	return found, last
 }
 
 // Count returns the number of subscriptions.
@@ -687,7 +909,10 @@ func (s *Sublist) Count() uint32 {
 
 // CacheCount returns the number of result sets in the cache.
 func (s *Sublist) CacheCount() int {
-	return int(atomic.LoadInt32(&s.cacheNum))
+	s.RLock()
+	cc := len(s.cache)
+	s.RUnlock()
+	return cc
 }
 
 // SublistStats are public stats for the sublist
@@ -702,6 +927,7 @@ type SublistStats struct {
 	AvgFanout    float64 `json:"avg_fanout"`
 	totFanout    int
 	cacheCnt     int
+	cacheHits    uint64
 }
 
 func (s *SublistStats) add(stat *SublistStats) {
@@ -710,7 +936,7 @@ func (s *SublistStats) add(stat *SublistStats) {
 	s.NumInserts += stat.NumInserts
 	s.NumRemoves += stat.NumRemoves
 	s.NumMatches += stat.NumMatches
-	s.CacheHitRate += stat.CacheHitRate
+	s.cacheHits += stat.cacheHits
 	if s.MaxFanout < stat.MaxFanout {
 		s.MaxFanout = stat.MaxFanout
 	}
@@ -722,6 +948,9 @@ func (s *SublistStats) add(stat *SublistStats) {
 	if s.totFanout > 0 {
 		s.AvgFanout = float64(s.totFanout) / float64(s.cacheCnt)
 	}
+	if s.NumMatches > 0 {
+		s.CacheHitRate = float64(s.cacheHits) / float64(s.NumMatches)
+	}
 }
 
 // Stats will return a stats structure for the current state.
@@ -730,33 +959,33 @@ func (s *Sublist) Stats() *SublistStats {
 
 	s.RLock()
 	cache := s.cache
+	cc := len(s.cache)
 	st.NumSubs = s.count
 	st.NumInserts = s.inserts
 	st.NumRemoves = s.removes
 	s.RUnlock()
 
-	if cn := atomic.LoadInt32(&s.cacheNum); cn > 0 {
-		st.NumCache = uint32(cn)
-	}
+	st.NumCache = uint32(cc)
 	st.NumMatches = atomic.LoadUint64(&s.matches)
+	st.cacheHits = atomic.LoadUint64(&s.cacheHits)
 	if st.NumMatches > 0 {
-		st.CacheHitRate = float64(atomic.LoadUint64(&s.cacheHits)) / float64(st.NumMatches)
+		st.CacheHitRate = float64(st.cacheHits) / float64(st.NumMatches)
 	}
 
 	// whip through cache for fanout stats, this can be off if cache is full and doing evictions.
 	// If this is called frequently, which it should not be, this could hurt performance.
 	if cache != nil {
 		tot, max, clen := 0, 0, 0
-		s.cache.Range(func(k, v interface{}) bool {
+		s.RLock()
+		for _, r := range s.cache {
 			clen++
-			r := v.(*SublistResult)
 			l := len(r.psubs) + len(r.qsubs)
 			tot += l
 			if l > max {
 				max = l
 			}
-			return true
-		})
+		}
+		s.RUnlock()
 		st.totFanout = tot
 		st.cacheCnt = clen
 		st.MaxFanout = uint32(max)
@@ -833,29 +1062,65 @@ func IsValidPublishSubject(subject string) bool {
 
 // IsValidSubject returns true if a subject is valid, false otherwise
 func IsValidSubject(subject string) bool {
-	if subject == "" {
+	if subject == _EMPTY_ {
 		return false
 	}
 	sfwc := false
 	tokens := strings.Split(subject, tsep)
 	for _, t := range tokens {
-		if len(t) == 0 || sfwc {
+		length := len(t)
+		if length == 0 || sfwc {
 			return false
 		}
+		if length > 1 {
+			if strings.ContainsAny(t, "\t\n\f\r ") {
+				return false
+			}
+			continue
+		}
+		switch t[0] {
+		case fwc:
+			sfwc = true
+		case ' ', '\t', '\n', '\r', '\f':
+			return false
+		}
+	}
+	return true
+}
+
+// Will share relevant info regarding the subject.
+// Returns valid, tokens, num pwcs, has fwc.
+func subjectInfo(subject string) (bool, []string, int, bool) {
+	if subject == "" {
+		return false, nil, 0, false
+	}
+	npwcs := 0
+	sfwc := false
+	tokens := strings.Split(subject, tsep)
+	for _, t := range tokens {
+		if len(t) == 0 || sfwc {
+			return false, nil, 0, false
+		}
 		if len(t) > 1 {
 			continue
 		}
 		switch t[0] {
 		case fwc:
 			sfwc = true
+		case pwc:
+			npwcs++
 		}
 	}
-	return true
+	return true, tokens, npwcs, sfwc
 }
 
 // IsValidLiteralSubject returns true if a subject is valid and literal (no wildcards), false otherwise
 func IsValidLiteralSubject(subject string) bool {
-	tokens := strings.Split(subject, tsep)
+	return isValidLiteralSubject(strings.Split(subject, tsep))
+}
+
+// isValidLiteralSubject returns true if the tokens are valid and literal (no wildcards), false otherwise
+func isValidLiteralSubject(tokens []string) bool {
 	for _, t := range tokens {
 		if len(t) == 0 {
 			return false
@@ -871,10 +1136,150 @@ func IsValidLiteralSubject(subject string) bool {
 	return true
 }
 
-// Calls into the function isSubsetMatch()
-func subjectIsSubsetMatch(subject, test string) bool {
-	tsa := [32]string{}
-	tts := tsa[:0]
+// ValidateMappingDestination returns nil error if the subject is a valid subject mapping destination subject
+func ValidateMappingDestination(subject string) error {
+	subjectTokens := strings.Split(subject, tsep)
+	sfwc := false
+	for _, t := range subjectTokens {
+		length := len(t)
+		if length == 0 || sfwc {
+			return &mappingDestinationErr{t, ErrInvalidMappingDestinationSubject}
+		}
+
+		if length > 4 && t[0] == '{' && t[1] == '{' && t[length-2] == '}' && t[length-1] == '}' {
+			if !partitionMappingFunctionRegEx.MatchString(t) &&
+				!wildcardMappingFunctionRegEx.MatchString(t) &&
+				!splitFromLeftMappingFunctionRegEx.MatchString(t) &&
+				!splitFromRightMappingFunctionRegEx.MatchString(t) &&
+				!sliceFromLeftMappingFunctionRegEx.MatchString(t) &&
+				!sliceFromRightMappingFunctionRegEx.MatchString(t) &&
+				!splitMappingFunctionRegEx.MatchString(t) {
+				return &mappingDestinationErr{t, ErrUnknownMappingDestinationFunction}
+			} else {
+				continue
+			}
+		}
+
+		if length == 1 && t[0] == fwc {
+			sfwc = true
+		} else if strings.ContainsAny(t, "\t\n\f\r ") {
+			return ErrInvalidMappingDestinationSubject
+		}
+	}
+	return nil
+}
+
+// Will check tokens and report back if the have any partial or full wildcards.
+func analyzeTokens(tokens []string) (hasPWC, hasFWC bool) {
+	for _, t := range tokens {
+		if lt := len(t); lt == 0 || lt > 1 {
+			continue
+		}
+		switch t[0] {
+		case pwc:
+			hasPWC = true
+		case fwc:
+			hasFWC = true
+		}
+	}
+	return
+}
+
+// Check on a token basis if they could match.
+func tokensCanMatch(t1, t2 string) bool {
+	if len(t1) == 0 || len(t2) == 0 {
+		return false
+	}
+	t1c, t2c := t1[0], t2[0]
+	if t1c == pwc || t2c == pwc || t1c == fwc || t2c == fwc {
+		return true
+	}
+	return t1 == t2
+}
+
+// SubjectsCollide will determine if two subjects could both match a single literal subject.
+func SubjectsCollide(subj1, subj2 string) bool {
+	if subj1 == subj2 {
+		return true
+	}
+	toks1 := strings.Split(subj1, tsep)
+	toks2 := strings.Split(subj2, tsep)
+	pwc1, fwc1 := analyzeTokens(toks1)
+	pwc2, fwc2 := analyzeTokens(toks2)
+	// if both literal just string compare.
+	l1, l2 := !(pwc1 || fwc1), !(pwc2 || fwc2)
+	if l1 && l2 {
+		return subj1 == subj2
+	}
+	// So one or both have wildcards. If one is literal than we can do subset matching.
+	if l1 && !l2 {
+		return isSubsetMatch(toks1, subj2)
+	} else if l2 && !l1 {
+		return isSubsetMatch(toks2, subj1)
+	}
+	// Both have wildcards.
+	// If they only have partials then the lengths must match.
+	if !fwc1 && !fwc2 && len(toks1) != len(toks2) {
+		return false
+	}
+	if lt1, lt2 := len(toks1), len(toks2); lt1 != lt2 {
+		// If the shorter one only has partials then these will not collide.
+		if lt1 < lt2 && !fwc1 || lt2 < lt1 && !fwc2 {
+			return false
+		}
+	}
+
+	stop := len(toks1)
+	if len(toks2) < stop {
+		stop = len(toks2)
+	}
+
+	// We look for reasons to say no.
+	for i := 0; i < stop; i++ {
+		t1, t2 := toks1[i], toks2[i]
+		if !tokensCanMatch(t1, t2) {
+			return false
+		}
+	}
+
+	return true
+}
+
+// Returns number of tokens in the subject.
+func numTokens(subject string) int {
+	var numTokens int
+	if len(subject) == 0 {
+		return 0
+	}
+	for i := 0; i < len(subject); i++ {
+		if subject[i] == btsep {
+			numTokens++
+		}
+	}
+	return numTokens + 1
+}
+
+// Fast way to return an indexed token.
+// This is one based, so first token is TokenAt(subject, 1)
+func tokenAt(subject string, index uint8) string {
+	ti, start := uint8(1), 0
+	for i := 0; i < len(subject); i++ {
+		if subject[i] == btsep {
+			if ti == index {
+				return subject[start:i]
+			}
+			start = i + 1
+			ti++
+		}
+	}
+	if ti == index {
+		return subject[start:]
+	}
+	return _EMPTY_
+}
+
+// use similar to append. meaning, the updated slice will be returned
+func tokenizeSubjectIntoSlice(tts []string, subject string) []string {
 	start := 0
 	for i := 0; i < len(subject); i++ {
 		if subject[i] == btsep {
@@ -883,27 +1288,31 @@ func subjectIsSubsetMatch(subject, test string) bool {
 		}
 	}
 	tts = append(tts, subject[start:])
+	return tts
+}
+
+// Calls into the function isSubsetMatch()
+func subjectIsSubsetMatch(subject, test string) bool {
+	tsa := [32]string{}
+	tts := tokenizeSubjectIntoSlice(tsa[:0], subject)
 	return isSubsetMatch(tts, test)
 }
 
 // This will test a subject as an array of tokens against a test subject
+// Calls into the function isSubsetMatchTokenized
+func isSubsetMatch(tokens []string, test string) bool {
+	tsa := [32]string{}
+	tts := tokenizeSubjectIntoSlice(tsa[:0], test)
+	return isSubsetMatchTokenized(tokens, tts)
+}
+
+// This will test a subject as an array of tokens against a test subject (also encoded as array of tokens)
 // and determine if the tokens are matched. Both test subject and tokens
 // may contain wildcards. So foo.* is a subset match of [">", "*.*", "foo.*"],
 // but not of foo.bar, etc.
-func isSubsetMatch(tokens []string, test string) bool {
-	tsa := [32]string{}
-	tts := tsa[:0]
-	start := 0
-	for i := 0; i < len(test); i++ {
-		if test[i] == btsep {
-			tts = append(tts, test[start:i])
-			start = i + 1
-		}
-	}
-	tts = append(tts, test[start:])
-
+func isSubsetMatchTokenized(tokens, test []string) bool {
 	// Walk the target tokens
-	for i, t2 := range tts {
+	for i, t2 := range test {
 		if i >= len(tokens) {
 			return false
 		}
@@ -915,19 +1324,18 @@ func isSubsetMatch(tokens []string, test string) bool {
 			return true
 		}
 		t1 := tokens[i]
+
 		l = len(t1)
-		if l == 0 {
-			return false
-		}
-		if t1[0] == fwc && l == 1 {
+		if l == 0 || t1[0] == fwc && l == 1 {
 			return false
 		}
+
 		if t1[0] == pwc && len(t1) == 1 {
 			m := t2[0] == pwc && len(t2) == 1
 			if !m {
 				return false
 			}
-			if i >= len(tts) {
+			if i >= len(test) {
 				return true
 			}
 			continue
@@ -936,7 +1344,7 @@ func isSubsetMatch(tokens []string, test string) bool {
 			return false
 		}
 	}
-	return len(tokens) == len(tts)
+	return len(tokens) == len(test)
 }
 
 // matchLiteral is used to test literal subjects, those that do not have any
@@ -1010,50 +1418,54 @@ func matchLiteral(literal, subject string) bool {
 	return li >= ll
 }
 
-func addLocalSub(sub *subscription, subs *[]*subscription) {
-	if sub != nil && sub.client != nil && (sub.client.kind == CLIENT || sub.client.kind == SYSTEM) && sub.im == nil {
-		*subs = append(*subs, sub)
+func addLocalSub(sub *subscription, subs *[]*subscription, includeLeafHubs bool) {
+	if sub != nil && sub.client != nil {
+		kind := sub.client.kind
+		if kind == CLIENT || kind == SYSTEM || kind == JETSTREAM || kind == ACCOUNT ||
+			(includeLeafHubs && sub.client.isHubLeafNode() /* implied kind==LEAF */) {
+			*subs = append(*subs, sub)
+		}
 	}
 }
 
-func (s *Sublist) addNodeToSubs(n *node, subs *[]*subscription) {
+func (s *Sublist) addNodeToSubs(n *node, subs *[]*subscription, includeLeafHubs bool) {
 	// Normal subscriptions
 	if n.plist != nil {
 		for _, sub := range n.plist {
-			addLocalSub(sub, subs)
+			addLocalSub(sub, subs, includeLeafHubs)
 		}
 	} else {
 		for _, sub := range n.psubs {
-			addLocalSub(sub, subs)
+			addLocalSub(sub, subs, includeLeafHubs)
 		}
 	}
 	// Queue subscriptions
 	for _, qr := range n.qsubs {
 		for _, sub := range qr {
-			addLocalSub(sub, subs)
+			addLocalSub(sub, subs, includeLeafHubs)
 		}
 	}
 }
 
-func (s *Sublist) collectLocalSubs(l *level, subs *[]*subscription) {
+func (s *Sublist) collectLocalSubs(l *level, subs *[]*subscription, includeLeafHubs bool) {
 	for _, n := range l.nodes {
-		s.addNodeToSubs(n, subs)
-		s.collectLocalSubs(n.next, subs)
+		s.addNodeToSubs(n, subs, includeLeafHubs)
+		s.collectLocalSubs(n.next, subs, includeLeafHubs)
 	}
 	if l.pwc != nil {
-		s.addNodeToSubs(l.pwc, subs)
-		s.collectLocalSubs(l.pwc.next, subs)
+		s.addNodeToSubs(l.pwc, subs, includeLeafHubs)
+		s.collectLocalSubs(l.pwc.next, subs, includeLeafHubs)
 	}
 	if l.fwc != nil {
-		s.addNodeToSubs(l.fwc, subs)
-		s.collectLocalSubs(l.fwc.next, subs)
+		s.addNodeToSubs(l.fwc, subs, includeLeafHubs)
+		s.collectLocalSubs(l.fwc.next, subs, includeLeafHubs)
 	}
 }
 
 // Return all local client subscriptions. Use the supplied slice.
-func (s *Sublist) localSubs(subs *[]*subscription) {
+func (s *Sublist) localSubs(subs *[]*subscription, includeLeafHubs bool) {
 	s.RLock()
-	s.collectLocalSubs(s.root, subs)
+	s.collectLocalSubs(s.root, subs, includeLeafHubs)
 	s.RUnlock()
 }
 
@@ -1096,16 +1508,77 @@ func (s *Sublist) collectAllSubs(l *level, subs *[]*subscription) {
 	}
 }
 
-// Helper to get the first result sub.
-func firstSubFromResult(rr *SublistResult) *subscription {
-	if rr == nil {
-		return nil
+// For a given subject (which may contain wildcards), this call returns all
+// subscriptions that would match that subject. For instance, suppose that
+// the sublist contains: foo.bar, foo.bar.baz and foo.baz, ReverseMatch("foo.*")
+// would return foo.bar and foo.baz.
+// This is used in situations where the sublist is likely to contain only
+// literals and one wants to get all the subjects that would have been a match
+// to a subscription on `subject`.
+func (s *Sublist) ReverseMatch(subject string) *SublistResult {
+	tsa := [32]string{}
+	tokens := tsa[:0]
+	start := 0
+	for i := 0; i < len(subject); i++ {
+		if subject[i] == btsep {
+			tokens = append(tokens, subject[start:i])
+			start = i + 1
+		}
 	}
-	if len(rr.psubs) > 0 {
-		return rr.psubs[0]
+	tokens = append(tokens, subject[start:])
+
+	result := &SublistResult{}
+
+	s.Lock()
+	reverseMatchLevel(s.root, tokens, nil, result)
+	// Check for empty result.
+	if len(result.psubs) == 0 && len(result.qsubs) == 0 {
+		result = emptyResult
+	}
+	s.Unlock()
+
+	return result
+}
+
+func reverseMatchLevel(l *level, toks []string, n *node, results *SublistResult) {
+	if l == nil {
+		return
+	}
+	for i, t := range toks {
+		if len(t) == 1 {
+			if t[0] == fwc {
+				getAllNodes(l, results)
+				return
+			} else if t[0] == pwc {
+				for _, n := range l.nodes {
+					reverseMatchLevel(n.next, toks[i+1:], n, results)
+				}
+				return
+			}
+		}
+		n = l.nodes[t]
+		if n == nil {
+			break
+		}
+		l = n.next
+	}
+	if n != nil {
+		addNodeToResults(n, results)
+	}
+}
+
+func getAllNodes(l *level, results *SublistResult) {
+	if l == nil {
+		return
+	}
+	if l.pwc != nil {
+		addNodeToResults(l.pwc, results)
+	}
+	if l.fwc != nil {
+		addNodeToResults(l.fwc, results)
+	}
+	for _, n := range l.nodes {
+		addNodeToResults(n, results)
+		getAllNodes(n.next, results)
 	}
-	if len(rr.qsubs) > 0 {
-		return rr.qsubs[0][0]
-	}
-	return nil
 }
