@@ -1,4 +1,4 @@
-// Copyright 2012-2019 The NATS Authors
+// Copyright 2012-2020 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -287,6 +287,75 @@ func createClientWithAccount(b *testing.B, account, host string, port int) net.C
 	return c
 }
 
+func benchOptionsForServiceImports() *server.Options {
+	o := DefaultTestOptions
+	o.Host = "127.0.0.1"
+	o.Port = -1
+	o.DisableShortFirstPing = true
+	foo := server.NewAccount("$foo")
+	bar := server.NewAccount("$bar")
+	o.Accounts = []*server.Account{foo, bar}
+
+	return &o
+}
+
+func addServiceImports(b *testing.B, s *server.Server) {
+	// Add a bunch of service exports with wildcards, similar to JS.
+	var exports = []string{
+		server.JSApiAccountInfo,
+		server.JSApiTemplateCreate,
+		server.JSApiTemplates,
+		server.JSApiTemplateInfo,
+		server.JSApiTemplateDelete,
+		server.JSApiStreamCreate,
+		server.JSApiStreamUpdate,
+		server.JSApiStreams,
+		server.JSApiStreamInfo,
+		server.JSApiStreamDelete,
+		server.JSApiStreamPurge,
+		server.JSApiMsgDelete,
+		server.JSApiConsumerCreate,
+		server.JSApiConsumers,
+		server.JSApiConsumerInfo,
+		server.JSApiConsumerDelete,
+	}
+	foo, _ := s.LookupAccount("$foo")
+	bar, _ := s.LookupAccount("$bar")
+
+	for _, export := range exports {
+		if err := bar.AddServiceExport(export, nil); err != nil {
+			b.Fatalf("Could not add service export: %v", err)
+		}
+		if err := foo.AddServiceImport(bar, export, ""); err != nil {
+			b.Fatalf("Could not add service import: %v", err)
+		}
+	}
+}
+
+func Benchmark__PubServiceImports(b *testing.B) {
+	o := benchOptionsForServiceImports()
+	s := RunServer(o)
+	defer s.Shutdown()
+
+	addServiceImports(b, s)
+
+	c := createClientWithAccount(b, "$foo", o.Host, o.Port)
+	defer c.Close()
+
+	bw := bufio.NewWriterSize(c, defaultSendBufSize)
+	sendOp := []byte("PUB foo 2\r\nok\r\n")
+
+	b.ResetTimer()
+	for i := 0; i < b.N; i++ {
+		bw.Write(sendOp)
+	}
+	err := bw.Flush()
+	if err != nil {
+		b.Errorf("Received error on FLUSH write: %v\n", err)
+	}
+	b.StopTimer()
+}
+
 func Benchmark___PubSubAccsImport(b *testing.B) {
 	o := benchDefaultOptionsForAccounts()
 	s := RunServer(o)
@@ -865,9 +934,10 @@ func doFanIn(b *testing.B, numServers, numPublishers, numSubscribers int, subjec
 	if b.N < numPublishers {
 		return
 	}
-	if numSubscribers > numPublishers {
-		b.Fatalf("Fan in tests should have numPublishers (%d) > numSubscribers (%d)", numPublishers, numSubscribers)
-	}
+	// Don't check for number of subscribers being lower than the number of publishers.
+	// We also use this bench to show the performance impact of increased number of publishers,
+	// and for those tests, the number of publishers will start at 1 and increase to 10,
+	// while the number of subscribers will always be 3.
 	if numSubscribers > 10 {
 		b.Fatalf("numSubscribers should be <= 10")
 	}
@@ -986,6 +1056,22 @@ func Benchmark___FanIn_128kx100x1(b *testing.B) {
 	doFanIn(b, 1, 100, 1, sub, sizedString(65536*2))
 }
 
+func Benchmark___BumpPubCount_1x3(b *testing.B) {
+	doFanIn(b, 1, 1, 3, sub, sizedString(128))
+}
+
+func Benchmark___BumpPubCount_2x3(b *testing.B) {
+	doFanIn(b, 1, 2, 3, sub, sizedString(128))
+}
+
+func Benchmark___BumpPubCount_5x3(b *testing.B) {
+	doFanIn(b, 1, 5, 3, sub, sizedString(128))
+}
+
+func Benchmark__BumpPubCount_10x3(b *testing.B) {
+	doFanIn(b, 1, 10, 3, sub, sizedString(128))
+}
+
 func testDefaultBenchOptionsForGateway(name string) *server.Options {
 	opts := testDefaultOptionsForGateway(name)
 	opts.DisableShortFirstPing = true
@@ -1305,92 +1391,3 @@ func Benchmark__GWs_Reqs_1_SubAll(b *testing.B) {
 func Benchmark__GWs_Reqs_1SubEach(b *testing.B) {
 	gatewaySendRequestsBench(b, false)
 }
-
-func Benchmark_____RoutedIntGraph(b *testing.B) {
-	s, o := RunServerWithConfig("./configs/srv_a.conf")
-	o.AllowNewAccounts = true
-	defer s.Shutdown()
-
-	numRoutes := 100
-	loop := b.N / numRoutes
-
-	type rh struct {
-		r      net.Conn
-		send   sendFun
-		expect expectFun
-		done   chan bool
-	}
-
-	routes := make([]*rh, 0, numRoutes)
-	for i := 0; i < numRoutes; i++ {
-		r := createRouteConn(b, o.Cluster.Host, o.Cluster.Port)
-		defer r.Close()
-
-		checkInfoMsg(b, r)
-		send, expect := setupRoute(b, r, o)
-		send("PING\r\n")
-		expect(pongRe)
-
-		bw := bufio.NewWriterSize(r, defaultSendBufSize)
-
-		account := fmt.Sprintf("$foo.account.%d", i)
-		for s := 0; s < loop; s++ {
-			bw.Write([]byte(fmt.Sprintf("RS+ %s foo.bar.%d\r\n", account, s)))
-		}
-		bw.Flush()
-		send("PING\r\n")
-		expect(pongRe)
-		routes = append(routes, &rh{r, send, expect, make(chan bool)})
-	}
-
-	startCh := make(chan bool)
-
-	unsubLoop := func(route *rh, ch chan bool, index int) {
-		bw := bufio.NewWriterSize(route.r, defaultSendBufSize)
-		account := fmt.Sprintf("$foo.account.%d", index)
-
-		// Wait for seed server's first PING so that it does
-		// not interfere with the expected PONG after sending
-		// all RS- and last PING.
-		// The wait here does not affect perf measurements
-		// since we do so *before* we signal that we are ready.
-		route.expect(pingRe)
-		route.send("PONG\r\n")
-
-		// Signal we are ready
-		close(ch)
-
-		// Wait to start up actual unsubs.
-		<-startCh
-
-		for i := 0; i < loop; i++ {
-			_, err := bw.Write([]byte(fmt.Sprintf("RS- %s foo.bar.%d\r\n", account, i)))
-			if err != nil {
-				b.Errorf("Received error on RS- write: %v\n", err)
-				return
-			}
-		}
-		err := bw.Flush()
-		if err != nil {
-			b.Errorf("Received error on FLUSH write: %v\n", err)
-			return
-		}
-		route.send("PING\r\n")
-		route.expect(pongRe)
-		close(route.done)
-	}
-
-	for i, route := range routes {
-		ch := make(chan bool)
-		go unsubLoop(route, ch, i)
-		<-ch
-	}
-
-	// Actual unsub test here.
-	b.ResetTimer()
-	close(startCh)
-	for _, route := range routes {
-		<-route.done
-	}
-	b.StopTimer()
-}
