@@ -1,4 +1,4 @@
-// Copyright 2018-2019 The NATS Authors
+// Copyright 2018-2022 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -50,6 +50,93 @@ func TestNewRouteInfoOnConnect(t *testing.T) {
 	if info.Nonce == "" {
 		t.Fatalf("Expected a non empty nonce in new route INFO")
 	}
+	// By default headers should be true.
+	if !info.Headers {
+		t.Fatalf("Expected to have headers on by default")
+	}
+	// Leafnode origin cluster support.
+	if !info.LNOC {
+		t.Fatalf("Expected to have leafnode origin cluster support")
+	}
+}
+
+func TestNewRouteHeaderSupport(t *testing.T) {
+	srvA, srvB, optsA, optsB := runServers(t)
+	defer srvA.Shutdown()
+	defer srvB.Shutdown()
+
+	clientA := createClientConn(t, optsA.Host, optsA.Port)
+	defer clientA.Close()
+
+	clientB := createClientConn(t, optsB.Host, optsB.Port)
+	defer clientB.Close()
+
+	sendA, expectA := setupHeaderConn(t, clientA)
+	sendA("SUB foo bar 22\r\n")
+	sendA("PING\r\n")
+	expectA(pongRe)
+
+	if err := checkExpectedSubs(1, srvA, srvB); err != nil {
+		t.Fatalf("%v", err)
+	}
+
+	sendB, expectB := setupHeaderConn(t, clientB)
+	// Can not have \r\n in payload fyi for regex.
+	sendB("HPUB foo reply 12 14\r\nK1:V1,K2:V2 ok\r\n")
+	sendB("PING\r\n")
+	expectB(pongRe)
+
+	expectHeaderMsgs := expectHeaderMsgsCommand(t, expectA)
+	matches := expectHeaderMsgs(1)
+	checkHmsg(t, matches[0], "foo", "22", "reply", "12", "14", "K1:V1,K2:V2 ", "ok")
+}
+
+func TestNewRouteHeaderSupportOldAndNew(t *testing.T) {
+	optsA := LoadConfig("./configs/srv_a.conf")
+	optsA.NoHeaderSupport = true
+	srvA := RunServer(optsA)
+	defer srvA.Shutdown()
+
+	srvB, optsB := RunServerWithConfig("./configs/srv_b.conf")
+	defer srvB.Shutdown()
+
+	checkClusterFormed(t, srvA, srvB)
+
+	clientA := createClientConn(t, optsA.Host, optsA.Port)
+	defer clientA.Close()
+
+	clientB := createClientConn(t, optsB.Host, optsB.Port)
+	defer clientB.Close()
+
+	sendA, expectA := setupHeaderConn(t, clientA)
+	sendA("SUB foo bar 22\r\n")
+	sendA("PING\r\n")
+	expectA(pongRe)
+
+	if err := checkExpectedSubs(1, srvA, srvB); err != nil {
+		t.Fatalf("%v", err)
+	}
+
+	sendB, expectB := setupHeaderConn(t, clientB)
+	// Can not have \r\n in payload fyi for regex.
+	sendB("HPUB foo reply 12 14\r\nK1:V1,K2:V2 ok\r\n")
+	sendB("PING\r\n")
+	expectB(pongRe)
+
+	expectMsgs := expectMsgsCommand(t, expectA)
+	matches := expectMsgs(1)
+	checkMsg(t, matches[0], "foo", "22", "reply", "2", "ok")
+}
+
+func sendRouteInfo(t *testing.T, rc net.Conn, routeSend sendFun, routeID string) {
+	info := checkInfoMsg(t, rc)
+	info.ID = routeID
+	info.Name = ""
+	b, err := json.Marshal(info)
+	if err != nil {
+		t.Fatalf("Could not marshal test route info: %v", err)
+	}
+	routeSend(fmt.Sprintf("INFO %s\r\n", b))
 }
 
 func TestNewRouteConnectSubs(t *testing.T) {
@@ -77,15 +164,7 @@ func TestNewRouteConnectSubs(t *testing.T) {
 	routeID := "RTEST_NEW:22"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
-
+	sendRouteInfo(t, rc, routeSend, routeID)
 	buf := routeExpect(rsubRe)
 
 	matches := rsubRe.FindAllSubmatch(buf, -1)
@@ -133,7 +212,7 @@ func TestNewRouteConnectSubsWithAccount(t *testing.T) {
 	c := createClientConn(t, opts.Host, opts.Port)
 	defer c.Close()
 
-	send, expect := setupConnWithAccount(t, c, accName)
+	send, expect := setupConnWithAccount(t, s, c, accName)
 
 	// Create 10 normal subs and 10 queue subscribers.
 	for i := 0; i < 10; i++ {
@@ -151,15 +230,7 @@ func TestNewRouteConnectSubsWithAccount(t *testing.T) {
 	routeID := "RTEST_NEW:22"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
-
+	sendRouteInfo(t, rc, routeSend, routeID)
 	buf := routeExpect(rsubRe)
 
 	matches := rsubRe.FindAllSubmatch(buf, -1)
@@ -212,7 +283,7 @@ func TestNewRouteRSubs(t *testing.T) {
 
 	// Create a client an account foo.
 	clientA := createClientConn(t, opts.Host, opts.Port)
-	sendA, expectA := setupConnWithAccount(t, clientA, "$foo")
+	sendA, expectA := setupConnWithAccount(t, s, clientA, "$foo")
 	defer clientA.Close()
 	sendA("PING\r\n")
 	expectA(pongRe)
@@ -231,13 +302,7 @@ func TestNewRouteRSubs(t *testing.T) {
 	routeID := "RTEST_NEW:33"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 	routeSend("PING\r\n")
 	routeExpect(pongRe)
 
@@ -247,7 +312,7 @@ func TestNewRouteRSubs(t *testing.T) {
 
 	// Now create a new client for account $bar and have them subscribe.
 	clientB := createClientConn(t, opts.Host, opts.Port)
-	sendB, expectB := setupConnWithAccount(t, clientB, "$bar")
+	sendB, expectB := setupConnWithAccount(t, s, clientB, "$bar")
 	defer clientB.Close()
 
 	sendB("PING\r\n")
@@ -309,13 +374,7 @@ func TestNewRouteProgressiveNormalSubs(t *testing.T) {
 	routeID := "RTEST_NEW:33"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 	routeSend("PING\r\n")
 	routeExpect(pongRe)
 
@@ -406,13 +465,7 @@ func TestNewRouteClientClosedWithNormalSubscriptions(t *testing.T) {
 	routeID := "RTEST_NEW:44"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 	routeSend("PING\r\n")
 	routeExpect(pongRe)
 
@@ -455,13 +508,7 @@ func TestNewRouteClientClosedWithQueueSubscriptions(t *testing.T) {
 	routeID := "RTEST_NEW:44"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 	routeSend("PING\r\n")
 	routeExpect(pongRe)
 
@@ -492,9 +539,6 @@ func TestNewRouteRUnsubAccountSpecific(t *testing.T) {
 	s, opts := runNewRouteServer(t)
 	defer s.Shutdown()
 
-	// Allow new accounts to be created on the fly.
-	opts.AllowNewAccounts = true
-
 	// Create a routeConn
 	rc := createRouteConn(t, opts.Cluster.Host, opts.Cluster.Port)
 	defer rc.Close()
@@ -502,17 +546,12 @@ func TestNewRouteRUnsubAccountSpecific(t *testing.T) {
 	routeID := "RTEST_NEW:77"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 
 	// Now create 500 subs on same subject but all different accounts.
 	for i := 0; i < 500; i++ {
 		account := fmt.Sprintf("$foo.account.%d", i)
+		s.RegisterAccount(account)
 		routeSend(fmt.Sprintf("RS+ %s foo\r\n", account))
 	}
 	routeSend("PING\r\n")
@@ -525,7 +564,7 @@ func TestNewRouteRUnsubAccountSpecific(t *testing.T) {
 	c := createClientConn(t, opts.Host, opts.Port)
 	defer c.Close()
 
-	send, expect := setupConnWithAccount(t, c, "$foo.account.22")
+	send, expect := setupConnWithAccount(t, s, c, "$foo.account.22")
 	send("PUB foo 2\r\nok\r\nPING\r\n")
 	expect(pongRe)
 	c.Close()
@@ -533,7 +572,7 @@ func TestNewRouteRUnsubAccountSpecific(t *testing.T) {
 	// But make sure we still receive on others
 	c = createClientConn(t, opts.Host, opts.Port)
 	defer c.Close()
-	send, expect = setupConnWithAccount(t, c, "$foo.account.33")
+	send, expect = setupConnWithAccount(t, s, c, "$foo.account.33")
 	send("PUB foo 2\r\nok\r\nPING\r\n")
 	expect(pongRe)
 
@@ -548,9 +587,6 @@ func TestNewRouteRSubCleanupOnDisconnect(t *testing.T) {
 	s, opts := runNewRouteServer(t)
 	defer s.Shutdown()
 
-	// Allow new accounts to be created on the fly.
-	opts.AllowNewAccounts = true
-
 	// Create a routeConn
 	rc := createRouteConn(t, opts.Cluster.Host, opts.Cluster.Port)
 	defer rc.Close()
@@ -558,13 +594,7 @@ func TestNewRouteRSubCleanupOnDisconnect(t *testing.T) {
 	routeID := "RTEST_NEW:77"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 
 	// Now create 100 subs on 3 different accounts.
 	for i := 0; i < 100; i++ {
@@ -596,13 +626,7 @@ func TestNewRouteSendSubsAndMsgs(t *testing.T) {
 	routeID := "RTEST_NEW:44"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 	routeSend("PING\r\n")
 	routeExpect(pongRe)
 
@@ -724,13 +748,7 @@ func TestNewRouteProcessRoutedMsgs(t *testing.T) {
 	routeID := "RTEST_NEW:55"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
 
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 	routeSend("PING\r\n")
 	routeExpect(pongRe)
 
@@ -753,7 +771,7 @@ func TestNewRouteProcessRoutedMsgs(t *testing.T) {
 	matches := expectMsgs(1)
 	checkMsg(t, matches[0], "foo", "1", "", "2", "ok")
 
-	// Now send in a RMSG to the route witha reply and make sure its delivered to the client.
+	// Now send in a RMSG to the route with a reply and make sure its delivered to the client.
 	routeSend("RMSG $G foo reply 2\r\nok\r\nPING\r\n")
 	routeExpect(pongRe)
 
@@ -862,15 +880,14 @@ func TestNewRouteSinglePublishOnNewAccount(t *testing.T) {
 	defer srvA.Shutdown()
 	defer srvB.Shutdown()
 
-	// Allow new accounts to be created on the fly.
-	optsA.AllowNewAccounts = true
-	optsB.AllowNewAccounts = true
+	srvA.RegisterAccount("$TEST22")
+	srvB.RegisterAccount("$TEST22")
 
 	// Create and establish a listener on foo for $TEST22 account.
 	clientA := createClientConn(t, optsA.Host, optsA.Port)
 	defer clientA.Close()
 
-	sendA, expectA := setupConnWithAccount(t, clientA, "$TEST22")
+	sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$TEST22")
 	sendA("SUB foo 1\r\nPING\r\n")
 	expectA(pongRe)
 
@@ -882,7 +899,7 @@ func TestNewRouteSinglePublishOnNewAccount(t *testing.T) {
 	defer clientB.Close()
 
 	// Send a message, flush to make sure server processed and close connection.
-	sendB, expectB := setupConnWithAccount(t, clientB, "$TEST22")
+	sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$TEST22")
 	sendB("PUB foo 2\r\nok\r\nPING\r\n")
 	expectB(pongRe)
 	clientB.Close()
@@ -898,15 +915,14 @@ func TestNewRouteSinglePublishToQueueSubscriberOnNewAccount(t *testing.T) {
 	defer srvA.Shutdown()
 	defer srvB.Shutdown()
 
-	// Allow new accounts to be created on the fly.
-	optsA.AllowNewAccounts = true
-	optsB.AllowNewAccounts = true
+	srvA.RegisterAccount("$TEST22")
+	srvB.RegisterAccount("$TEST22")
 
 	// Create and establish a listener on foo for $TEST22 account.
 	clientA := createClientConn(t, optsA.Host, optsA.Port)
 	defer clientA.Close()
 
-	sendA, expectA := setupConnWithAccount(t, clientA, "$TEST22")
+	sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$TEST22")
 	sendA("SUB foo bar 1\r\nPING\r\n")
 	expectA(pongRe)
 
@@ -914,7 +930,7 @@ func TestNewRouteSinglePublishToQueueSubscriberOnNewAccount(t *testing.T) {
 	defer clientB.Close()
 
 	// Send a message, flush to make sure server processed and close connection.
-	sendB, expectB := setupConnWithAccount(t, clientB, "$TEST22")
+	sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$TEST22")
 	sendB("PUB foo bar 2\r\nok\r\nPING\r\n")
 	expectB(pongRe)
 	defer clientB.Close()
@@ -936,23 +952,22 @@ func TestNewRouteSinglePublishToMultipleQueueSubscriberOnNewAccount(t *testing.T
 	defer srvB.Shutdown()
 	defer srvC.Shutdown()
 
-	// Allow new accounts to be created on the fly.
-	optsA.AllowNewAccounts = true
-	optsB.AllowNewAccounts = true
-	optsC.AllowNewAccounts = true
+	srvA.RegisterAccount("$TEST22")
+	srvB.RegisterAccount("$TEST22")
+	srvC.RegisterAccount("$TEST22")
 
 	// Create and establish a listener on foo/bar for $TEST22 account. Do this on ClientA and ClientC.
 	clientA := createClientConn(t, optsA.Host, optsA.Port)
 	defer clientA.Close()
 
-	sendA, expectA := setupConnWithAccount(t, clientA, "$TEST22")
+	sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$TEST22")
 	sendA("SUB foo bar 11\r\nPING\r\n")
 	expectA(pongRe)
 
 	clientC := createClientConn(t, optsC.Host, optsC.Port)
 	defer clientC.Close()
 
-	sendC, expectC := setupConnWithAccount(t, clientC, "$TEST22")
+	sendC, expectC := setupConnWithAccount(t, srvC, clientC, "$TEST22")
 	sendC("SUB foo bar 33\r\nPING\r\n")
 	expectC(pongRe)
 
@@ -967,7 +982,7 @@ func TestNewRouteSinglePublishToMultipleQueueSubscriberOnNewAccount(t *testing.T
 	time.Sleep(100 * time.Millisecond)
 
 	// Send a message, flush to make sure server processed and close connection.
-	sendB, expectB := setupConnWithAccount(t, clientB, "$TEST22")
+	sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$TEST22")
 	sendB("PUB foo 2\r\nok\r\nPING\r\n")
 	expectB(pongRe)
 	defer clientB.Close()
@@ -1063,14 +1078,14 @@ func testNewRouteStreamImport(t *testing.T, duplicateSub bool) {
 	clientA := createClientConn(t, optsA.Host, optsA.Port)
 	defer clientA.Close()
 
-	sendA, expectA := setupConnWithAccount(t, clientA, "$foo")
+	sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$foo")
 
 	// Now setup client B on srvB who will do a sub from account $bar
 	// that should map account $foo's foo subject.
 	clientB := createClientConn(t, optsB.Host, optsB.Port)
 	defer clientB.Close()
 
-	sendB, expectB := setupConnWithAccount(t, clientB, "$bar")
+	sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$bar")
 	sendB("SUB foo 1\r\n")
 	if duplicateSub {
 		sendB("SUB foo 1\r\n")
@@ -1160,7 +1175,7 @@ func TestNewRouteStreamImportLargeFanout(t *testing.T) {
 	for i := 0; i < fanout; i++ {
 		clientB[i] = createClientConn(t, optsB.Host, optsB.Port)
 		defer clientB[i].Close()
-		sendB[i], expectB[i] = setupConnWithAccount(t, clientB[i], barA[i].Name)
+		sendB[i], expectB[i] = setupConnWithAccount(t, srvB, clientB[i], barA[i].Name)
 		sendB[i]("SUB foo 1\r\nPING\r\n")
 		expectB[i](pongRe)
 	}
@@ -1229,7 +1244,7 @@ func TestNewRouteServiceImport(t *testing.T) {
 	clientA := createClientConn(t, optsA.Host, optsA.Port)
 	defer clientA.Close()
 
-	sendA, expectA := setupConnWithAccount(t, clientA, "$foo")
+	sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$foo")
 	sendA("SUB test.request 1\r\nPING\r\n")
 	expectA(pongRe)
 
@@ -1238,12 +1253,12 @@ func TestNewRouteServiceImport(t *testing.T) {
 	clientB := createClientConn(t, optsB.Host, optsB.Port)
 	defer clientB.Close()
 
-	sendB, expectB := setupConnWithAccount(t, clientB, "$bar")
+	sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$bar")
 	sendB("SUB reply 1\r\nPING\r\n")
 	expectB(pongRe)
 
-	// Wait for all subs to be propagated.
-	if err := checkExpectedSubs(2, srvA, srvB); err != nil {
+	// Wait for all subs to be propagated. (1 on foo, 2 on bar)
+	if err := checkExpectedSubs(3, srvA, srvB); err != nil {
 		t.Fatal(err.Error())
 	}
 
@@ -1330,7 +1345,7 @@ func TestNewRouteServiceExportWithWildcards(t *testing.T) {
 			clientA := createClientConn(t, optsA.Host, optsA.Port)
 			defer clientA.Close()
 
-			sendA, expectA := setupConnWithAccount(t, clientA, "$foo")
+			sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$foo")
 			sendA("SUB ngs.update.* 1\r\nPING\r\n")
 			expectA(pongRe)
 
@@ -1339,12 +1354,12 @@ func TestNewRouteServiceExportWithWildcards(t *testing.T) {
 			clientB := createClientConn(t, optsB.Host, optsB.Port)
 			defer clientB.Close()
 
-			sendB, expectB := setupConnWithAccount(t, clientB, "$bar")
+			sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$bar")
 			sendB("SUB reply 1\r\nPING\r\n")
 			expectB(pongRe)
 
-			// Wait for all subs to be propagated
-			if err := checkExpectedSubs(2, srvA, srvB); err != nil {
+			// Wait for all subs to be propagated. (1 on foo, 2 on bar)
+			if err := checkExpectedSubs(3, srvA, srvB); err != nil {
 				t.Fatal(err.Error())
 			}
 
@@ -1410,7 +1425,7 @@ func TestNewRouteServiceImportQueueGroups(t *testing.T) {
 	clientA := createClientConn(t, optsA.Host, optsA.Port)
 	defer clientA.Close()
 
-	sendA, expectA := setupConnWithAccount(t, clientA, "$foo")
+	sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$foo")
 	sendA("SUB test.request QGROUP 1\r\nPING\r\n")
 	expectA(pongRe)
 
@@ -1419,12 +1434,12 @@ func TestNewRouteServiceImportQueueGroups(t *testing.T) {
 	clientB := createClientConn(t, optsB.Host, optsB.Port)
 	defer clientB.Close()
 
-	sendB, expectB := setupConnWithAccount(t, clientB, "$bar")
+	sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$bar")
 	sendB("SUB reply QGROUP_TOO 1\r\nPING\r\n")
 	expectB(pongRe)
 
-	// Wait for all subs to be propagated.
-	if err := checkExpectedSubs(2, srvA, srvB); err != nil {
+	// Wait for all subs to be propagated. (1 on foo, 2 on bar)
+	if err := checkExpectedSubs(3, srvA, srvB); err != nil {
 		t.Fatal(err.Error())
 	}
 
@@ -1472,8 +1487,6 @@ func TestNewRouteServiceImportDanglingRemoteSubs(t *testing.T) {
 	fooA, _ := registerAccounts(t, srvA)
 	fooB, barB := registerAccounts(t, srvB)
 
-	fooA.SetAutoExpireTTL(10 * time.Millisecond)
-
 	// Add in the service export for the requests. Make it public.
 	if err := fooA.AddServiceExport("test.request", nil); err != nil {
 		t.Fatalf("Error adding account service export to client foo: %v", err)
@@ -1491,7 +1504,7 @@ func TestNewRouteServiceImportDanglingRemoteSubs(t *testing.T) {
 	clientA := createClientConn(t, optsA.Host, optsA.Port)
 	defer clientA.Close()
 
-	sendA, expectA := setupConnWithAccount(t, clientA, "$foo")
+	sendA, expectA := setupConnWithAccount(t, srvA, clientA, "$foo")
 	// Express interest.
 	sendA("SUB test.request 1\r\nPING\r\n")
 	expectA(pongRe)
@@ -1501,7 +1514,7 @@ func TestNewRouteServiceImportDanglingRemoteSubs(t *testing.T) {
 	clientB := createClientConn(t, optsB.Host, optsB.Port)
 	defer clientB.Close()
 
-	sendB, expectB := setupConnWithAccount(t, clientB, "$bar")
+	sendB, expectB := setupConnWithAccount(t, srvB, clientB, "$bar")
 	sendB("SUB reply 1\r\nPING\r\n")
 	expectB(pongRe)
 
@@ -1510,8 +1523,8 @@ func TestNewRouteServiceImportDanglingRemoteSubs(t *testing.T) {
 	if err := checkExpectedSubs(2, srvA); err != nil {
 		t.Fatal(err.Error())
 	}
-	// Wait for all subs to be propagated
-	if err := checkExpectedSubs(2, srvB); err != nil {
+	// Wait for all subs to be propagated (1 on foo and 2 on bar)
+	if err := checkExpectedSubs(3, srvB); err != nil {
 		t.Fatal(err.Error())
 	}
 
@@ -1655,6 +1668,118 @@ func TestNewRouteLargeDistinctQueueSubscribers(t *testing.T) {
 	})
 }
 
+func TestNewRouteLeafNodeOriginSupport(t *testing.T) {
+	content := `
+	listen: 127.0.0.1:-1
+	cluster { name: xyz, listen: 127.0.0.1:-1 }
+	leafnodes { listen: 127.0.0.1:-1 }
+	no_sys_acc: true
+	`
+	conf := createConfFile(t, []byte(content))
+
+	s, opts := RunServerWithConfig(conf)
+	defer s.Shutdown()
+
+	gacc, _ := s.LookupAccount("$G")
+
+	lcontent := `
+	listen: 127.0.0.1:-1
+	cluster { name: ln1, listen: 127.0.0.1:-1 }
+	leafnodes { remotes = [{ url: nats-leaf://127.0.0.1:%d }] }
+	no_sys_acc: true
+	`
+	lconf := createConfFile(t, []byte(fmt.Sprintf(lcontent, opts.LeafNode.Port)))
+
+	ln, _ := RunServerWithConfig(lconf)
+	defer ln.Shutdown()
+
+	checkLeafNodeConnected(t, s)
+
+	lgacc, _ := ln.LookupAccount("$G")
+
+	rc := createRouteConn(t, opts.Cluster.Host, opts.Cluster.Port)
+	defer rc.Close()
+
+	routeID := "LNOC:22"
+	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
+
+	pingPong := func() {
+		t.Helper()
+		routeSend("PING\r\n")
+		routeExpect(pongRe)
+	}
+
+	info := checkInfoMsg(t, rc)
+	info.ID = routeID
+	info.Name = ""
+	info.LNOC = true
+	b, err := json.Marshal(info)
+	if err != nil {
+		t.Fatalf("Could not marshal test route info: %v", err)
+	}
+
+	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	routeExpect(rsubRe)
+	pingPong()
+
+	// Make sure it can process and LS+
+	routeSend("LS+ ln1 $G foo\r\n")
+	pingPong()
+
+	if !gacc.SubscriptionInterest("foo") {
+		t.Fatalf("Expected interest on \"foo\"")
+	}
+
+	// This should not have been sent to the leafnode since same origin cluster.
+	time.Sleep(10 * time.Millisecond)
+	if lgacc.SubscriptionInterest("foo") {
+		t.Fatalf("Did not expect interest on \"foo\"")
+	}
+
+	// Create a connection on the leafnode server.
+	nc, err := nats.Connect(ln.ClientURL())
+	if err != nil {
+		t.Fatalf("Unexpected error connecting %v", err)
+	}
+	defer nc.Close()
+
+	sub, _ := nc.SubscribeSync("bar")
+	// Let it propagate to the main server
+	checkFor(t, time.Second, 10*time.Millisecond, func() error {
+		if !gacc.SubscriptionInterest("bar") {
+			return fmt.Errorf("No interest")
+		}
+		return nil
+	})
+	// For "bar"
+	routeExpect(rlsubRe)
+
+	// Now pretend like we send a message to the main server over the
+	// route but from the same origin cluster, should not be delivered
+	// to the leafnode.
+
+	// Make sure it can process and LMSG.
+	// LMSG for routes is like HMSG with an origin cluster before the account.
+	routeSend("LMSG ln1 $G bar 0 2\r\nok\r\n")
+	pingPong()
+
+	// Let it propagate if not properly truncated.
+	time.Sleep(10 * time.Millisecond)
+	if n, _, _ := sub.Pending(); n != 0 {
+		t.Fatalf("Should not have received the message on bar")
+	}
+
+	// Try one with all the bells and whistles.
+	routeSend("LMSG ln1 $G foo + reply bar baz 0 2\r\nok\r\n")
+	pingPong()
+
+	// Let it propagate if not properly truncated.
+	time.Sleep(10 * time.Millisecond)
+	if n, _, _ := sub.Pending(); n != 0 {
+		t.Fatalf("Should not have received the message on bar")
+	}
+}
+
 // Check that real duplicate subscription (that is, sent by client with same sid)
 // are ignored and do not register multiple shadow subscriptions.
 func TestNewRouteDuplicateSubscription(t *testing.T) {
@@ -1671,13 +1796,7 @@ func TestNewRouteDuplicateSubscription(t *testing.T) {
 
 	routeID := "RTEST_DUPLICATE:22"
 	routeSend, routeExpect := setupRouteEx(t, rc, opts, routeID)
-	info := checkInfoMsg(t, rc)
-	info.ID = routeID
-	b, err := json.Marshal(info)
-	if err != nil {
-		t.Fatalf("Could not marshal test route info: %v", err)
-	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	sendRouteInfo(t, rc, routeSend, routeID)
 	routeSend("PING\r\n")
 	routeExpect(pongRe)
 
