@@ -20,8 +20,10 @@ import (
 	"fmt"
 	"io"
 	"net"
+	"os"
 	"regexp"
 	"runtime"
+	"strconv"
 	"strings"
 	"testing"
 	"time"
@@ -41,7 +43,7 @@ var DefaultTestOptions = server.Options{
 	Port:                  4222,
 	NoLog:                 true,
 	NoSigs:                true,
-	MaxControlLine:        2048,
+	MaxControlLine:        4096,
 	DisableShortFirstPing: true,
 }
 
@@ -50,6 +52,12 @@ func RunDefaultServer() *server.Server {
 	return RunServer(&DefaultTestOptions)
 }
 
+func RunRandClientPortServer() *server.Server {
+	opts := DefaultTestOptions
+	opts.Port = -1
+	return RunServer(&opts)
+}
+
 // To turn on server tracing and debugging and logging which are
 // normally suppressed.
 var (
@@ -60,6 +68,10 @@ var (
 
 // RunServer starts a new Go routine based server
 func RunServer(opts *server.Options) *server.Server {
+	return RunServerCallback(opts, nil)
+}
+
+func RunServerCallback(opts *server.Options, callback func(*server.Server)) *server.Server {
 	if opts == nil {
 		opts = &DefaultTestOptions
 	}
@@ -77,6 +89,10 @@ func RunServer(opts *server.Options) *server.Server {
 		s.ConfigureLogger()
 	}
 
+	if callback != nil {
+		callback(s)
+	}
+
 	// Run server in Go routine.
 	go s.Start()
 
@@ -103,6 +119,17 @@ func RunServerWithConfig(configFile string) (srv *server.Server, opts *server.Op
 	return
 }
 
+// RunServerWithConfigOverrides starts a new Go routine based server with a configuration file,
+// providing a callback to update the options configured.
+func RunServerWithConfigOverrides(configFile string, optsCallback func(*server.Options), svrCallback func(*server.Server)) (srv *server.Server, opts *server.Options) {
+	opts = LoadConfig(configFile)
+	if optsCallback != nil {
+		optsCallback(opts)
+	}
+	srv = RunServerCallback(opts, svrCallback)
+	return
+}
+
 func stackFatalf(t tLogger, f string, args ...interface{}) {
 	lines := make([]string, 0, 32)
 	msg := fmt.Sprintf(f, args...)
@@ -189,21 +216,30 @@ func checkInfoMsg(t tLogger, c net.Conn) server.Info {
 	return sinfo
 }
 
-func doConnect(t tLogger, c net.Conn, verbose, pedantic, ssl bool) {
+func doHeadersConnect(t tLogger, c net.Conn, verbose, pedantic, ssl, headers bool) {
 	checkInfoMsg(t, c)
-	cs := fmt.Sprintf("CONNECT {\"verbose\":%v,\"pedantic\":%v,\"tls_required\":%v}\r\n", verbose, pedantic, ssl)
+	cs := fmt.Sprintf("CONNECT {\"verbose\":%v,\"pedantic\":%v,\"tls_required\":%v,\"headers\":%v}\r\n",
+		verbose, pedantic, ssl, headers)
 	sendProto(t, c, cs)
 }
 
+func doConnect(t tLogger, c net.Conn, verbose, pedantic, ssl bool) {
+	doHeadersConnect(t, c, verbose, pedantic, ssl, false)
+}
+
+func doDefaultHeadersConnect(t tLogger, c net.Conn) {
+	doHeadersConnect(t, c, false, false, false, true)
+}
+
 func doDefaultConnect(t tLogger, c net.Conn) {
 	// Basic Connect
 	doConnect(t, c, false, false, false)
 }
 
-const connectProto = "CONNECT {\"verbose\":false,\"user\":\"%s\",\"pass\":\"%s\",\"name\":\"%s\"}\r\n"
+const routeConnectProto = "CONNECT {\"verbose\":false,\"user\":\"%s\",\"pass\":\"%s\",\"name\":\"%s\",\"cluster\":\"xyz\"}\r\n"
 
 func doRouteAuthConnect(t tLogger, c net.Conn, user, pass, id string) {
-	cs := fmt.Sprintf(connectProto, user, pass, id)
+	cs := fmt.Sprintf(routeConnectProto, user, pass, id)
 	sendProto(t, c, cs)
 }
 
@@ -221,6 +257,11 @@ func setupRoute(t tLogger, c net.Conn, opts *server.Options) (sendFun, expectFun
 	return setupRouteEx(t, c, opts, id)
 }
 
+func setupHeaderConn(t tLogger, c net.Conn) (sendFun, expectFun) {
+	doDefaultHeadersConnect(t, c)
+	return sendCommand(t, c), expectCommand(t, c)
+}
+
 func setupConn(t tLogger, c net.Conn) (sendFun, expectFun) {
 	doDefaultConnect(t, c)
 	return sendCommand(t, c), expectCommand(t, c)
@@ -233,11 +274,27 @@ func setupConnWithProto(t tLogger, c net.Conn, proto int) (sendFun, expectFun) {
 	return sendCommand(t, c), expectCommand(t, c)
 }
 
-func setupConnWithAccount(t tLogger, c net.Conn, account string) (sendFun, expectFun) {
-	checkInfoMsg(t, c)
-	cs := fmt.Sprintf("CONNECT {\"verbose\":%v,\"pedantic\":%v,\"tls_required\":%v,\"account\":%q}\r\n", false, false, false, account)
+func setupConnWithAccount(t tLogger, s *server.Server, c net.Conn, account string) (sendFun, expectFun) {
+	info := checkInfoMsg(t, c)
+	s.RegisterAccount(account)
+	acc, err := s.LookupAccount(account)
+	if err != nil {
+		t.Fatalf("Unexpected Error: %v", err)
+	}
+	cs := fmt.Sprintf("CONNECT {\"verbose\":%v,\"pedantic\":%v,\"tls_required\":%v}\r\n", false, false, false)
 	sendProto(t, c, cs)
-	return sendCommand(t, c), expectCommand(t, c)
+
+	send, expect := sendCommand(t, c), expectCommand(t, c)
+	send("PING\r\n")
+	expect(pongRe)
+
+	nc := s.GetClient(info.CID)
+	if nc == nil {
+		t.Fatalf("Could not get client for CID:%d", info.CID)
+	}
+	nc.RegisterUser(&server.User{Account: acc})
+
+	return send, expect
 }
 
 func setupConnWithUserPass(t tLogger, c net.Conn, username, password string) (sendFun, expectFun) {
@@ -245,7 +302,7 @@ func setupConnWithUserPass(t tLogger, c net.Conn, username, password string) (se
 	cs := fmt.Sprintf("CONNECT {\"verbose\":%v,\"pedantic\":%v,\"tls_required\":%v,\"protocol\":1,\"user\":%q,\"pass\":%q}\r\n",
 		false, false, false, username, password)
 	sendProto(t, c, cs)
-	return sendCommand(t, c), expectCommand(t, c)
+	return sendCommand(t, c), expectLefMostCommand(t, c)
 }
 
 type sendFun func(string)
@@ -265,6 +322,14 @@ func expectCommand(t tLogger, c net.Conn) expectFun {
 	}
 }
 
+// Closure version for easier reading
+func expectLefMostCommand(t tLogger, c net.Conn) expectFun {
+	var buf []byte
+	return func(re *regexp.Regexp) []byte {
+		return expectLeftMostResult(t, c, re, &buf)
+	}
+}
+
 // Send the protocol command to the server.
 func sendProto(t tLogger, c net.Conn, op string) {
 	n, err := c.Write([]byte(op))
@@ -277,9 +342,11 @@ func sendProto(t tLogger, c net.Conn, op string) {
 }
 
 var (
+	anyRe     = regexp.MustCompile(`.*`)
 	infoRe    = regexp.MustCompile(`INFO\s+([^\r\n]+)\r\n`)
 	pingRe    = regexp.MustCompile(`^PING\r\n`)
 	pongRe    = regexp.MustCompile(`^PONG\r\n`)
+	hmsgRe    = regexp.MustCompile(`(?:(?:HMSG\s+([^\s]+)\s+([^\s]+)\s+(([^\s]+)[^\S\r\n]+)?(\d+)\s+(\d+)\s*\r\n([^\\r\\n]*?)\r\n)+?)`)
 	msgRe     = regexp.MustCompile(`(?:(?:MSG\s+([^\s]+)\s+([^\s]+)\s+(([^\s]+)[^\S\r\n]+)?(\d+)\s*\r\n([^\\r\\n]*?)\r\n)+?)`)
 	rawMsgRe  = regexp.MustCompile(`(?:(?:MSG\s+([^\s]+)\s+([^\s]+)\s+(([^\s]+)[^\S\r\n]+)?(\d+)\s*\r\n(.*?)))`)
 	okRe      = regexp.MustCompile(`\A\+OK\r\n`)
@@ -293,6 +360,7 @@ var (
 	lsubRe    = regexp.MustCompile(`LS\+\s+([^\s]+)\s*([^\s]+)?\s*(\d+)?\r\n`)
 	lunsubRe  = regexp.MustCompile(`LS\-\s+([^\s]+)\s*([^\s]+)?\r\n`)
 	lmsgRe    = regexp.MustCompile(`(?:(?:LMSG\s+([^\s]+)\s+(?:([|+]\s+([\w\s]+)|[^\s]+)[^\S\r\n]+)?(\d+)\s*\r\n([^\\r\\n]*?)\r\n)+?)`)
+	rlsubRe   = regexp.MustCompile(`LS\+\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s*([^\s]+)?\s*(\d+)?\r\n`)
 )
 
 const (
@@ -302,6 +370,10 @@ const (
 	replyIndex = 4
 	lenIndex   = 5
 	msgIndex   = 6
+	// Headers
+	hlenIndex = 5
+	tlenIndex = 6
+	hmsgIndex = 7
 
 	// Routed Messages
 	accIndex           = 1
@@ -309,6 +381,41 @@ const (
 	replyAndQueueIndex = 3
 )
 
+// Test result from server against regexp and return left most match
+func expectLeftMostResult(t tLogger, c net.Conn, re *regexp.Regexp, buf *[]byte) []byte {
+	recv := func() []byte {
+		expBuf := make([]byte, 32768)
+		// Wait for commands to be processed and results queued for read
+		c.SetReadDeadline(time.Now().Add(2 * time.Second))
+		n, err := c.Read(expBuf)
+		c.SetReadDeadline(time.Time{})
+
+		if n <= 0 && err != nil {
+			stackFatalf(t, "Error reading from conn: %v\n", err)
+		}
+		return expBuf[:n]
+	}
+	if len(*buf) == 0 {
+		*buf = recv()
+	}
+	emptyCnt := 0
+	for {
+		result := re.Find(*buf)
+		if result == nil {
+			emptyCnt++
+			if emptyCnt > 5 {
+				stackFatalf(t, "Reading empty data too often\n")
+			}
+			*buf = append(*buf, recv()...)
+		} else {
+			emptyCnt = 0
+			cutIdx := strings.Index(string(*buf), string(result)) + len(result)
+			*buf = (*buf)[cutIdx:]
+			return result
+		}
+	}
+}
+
 // Test result from server against regexp
 func expectResult(t tLogger, c net.Conn, re *regexp.Regexp) []byte {
 	expBuf := make([]byte, 32768)
@@ -321,7 +428,6 @@ func expectResult(t tLogger, c net.Conn, re *regexp.Regexp) []byte {
 		stackFatalf(t, "Error reading from conn: %v\n", err)
 	}
 	buf := expBuf[:n]
-
 	if !re.Match(buf) {
 		stackFatalf(t, "Response did not match expected: \n\tReceived:'%q'\n\tExpected:'%s'", buf, re)
 	}
@@ -410,6 +516,35 @@ func checkLmsg(t tLogger, m [][]byte, subject, replyAndQueues, len, msg string)
 	}
 }
 
+// This will check that we got what we expected from a header message.
+func checkHmsg(t tLogger, m [][]byte, subject, sid, reply, hlen, len, hdr, msg string) {
+	if string(m[subIndex]) != subject {
+		stackFatalf(t, "Did not get correct subject: expected '%s' got '%s'\n", subject, m[subIndex])
+	}
+	if sid != "" && string(m[sidIndex]) != sid {
+		stackFatalf(t, "Did not get correct sid: expected '%s' got '%s'\n", sid, m[sidIndex])
+	}
+	if string(m[replyIndex]) != reply {
+		stackFatalf(t, "Did not get correct reply: expected '%s' got '%s'\n", reply, m[replyIndex])
+	}
+	if string(m[hlenIndex]) != hlen {
+		stackFatalf(t, "Did not get correct header length: expected '%s' got '%s'\n", hlen, m[hlenIndex])
+	}
+	if string(m[tlenIndex]) != len {
+		stackFatalf(t, "Did not get correct msg length: expected '%s' got '%s'\n", len, m[tlenIndex])
+	}
+	// Extract the payload and break up the headers and msg.
+	payload := string(m[hmsgIndex])
+	hi, _ := strconv.Atoi(hlen)
+	rhdr, rmsg := payload[:hi], payload[hi:]
+	if rhdr != hdr {
+		stackFatalf(t, "Did not get correct headers: expected '%s' got '%s'\n", hdr, rhdr)
+	}
+	if rmsg != msg {
+		stackFatalf(t, "Did not get correct msg: expected '%s' got '%s'\n", msg, rmsg)
+	}
+}
+
 // Closure for expectMsgs
 func expectRmsgsCommand(t tLogger, ef expectFun) func(int) [][][]byte {
 	return func(expected int) [][][]byte {
@@ -422,6 +557,18 @@ func expectRmsgsCommand(t tLogger, ef expectFun) func(int) [][][]byte {
 	}
 }
 
+// Closure for expectHMsgs
+func expectHeaderMsgsCommand(t tLogger, ef expectFun) func(int) [][][]byte {
+	return func(expected int) [][][]byte {
+		buf := ef(hmsgRe)
+		matches := hmsgRe.FindAllSubmatch(buf, -1)
+		if len(matches) != expected {
+			stackFatalf(t, "Did not get correct # msgs: %d vs %d\n", len(matches), expected)
+		}
+		return matches
+	}
+}
+
 // Closure for expectMsgs
 func expectMsgsCommand(t tLogger, ef expectFun) func(int) [][][]byte {
 	return func(expected int) [][][]byte {
@@ -487,3 +634,12 @@ func nextServerOpts(opts *server.Options) *server.Options {
 	nopts.HTTPPort++
 	return nopts
 }
+
+func createTempFile(t testing.TB, prefix string) *os.File {
+	t.Helper()
+	file, err := os.CreateTemp(t.TempDir(), prefix)
+	if err != nil {
+		t.Fatal(err)
+	}
+	return file
+}
